from optparse import OptionGroup, Option
from hachoir.plugin import printParserList
from hachoir.stream import FileInputStream
from hachoir.error import error, warning, info
from hachoir.log import log
from hachoir.plugin import guessParser, loadParserPlugins
from hachoir.tools import humanDurationNanosec
import hachoir.config as config
import locale
import math
import operator
import sys
import time

def getTerminalCharset(default_charset="ascii"):
    """
    Guess terminal charset using differents tests:
    1. Try locale.getpreferredencoding()
    2. Try locale.nl_langinfo(CODESET)
    3. Try sys.stdout.encoding
    4. Otherwise, returns default charset

    WARNING: Call following code before calling this function:
    >>> import locale
    >>> locale.setlocale(locale.LC_ALL, "")
    """

    # (1) Try locale.getpreferredencoding()
    try:
        return locale.getpreferredencoding()
    except (locale.Error, AttributeError):
        pass

    # (2) Try locale.nl_langinfo(CODESET)
    try:
        return locale.nl_langinfo(locale.CODESET)
    except (locale.Error, AttributeError):
        pass

    # (3) Try sys.stdout.encoding
    if hasattr(sys.stdout, "encoding"):
        return sys.stdout.encoding

    # (4) Otherwise, returns default charset
    return default_charset

def displayParserList(*args):
    printParserList()
    sys.exit(0)

def _setLogFilename(*args):
    log.setFilename(args[2])

def getHachoirOptions(parser, console_ui=False):
    """
    Create an option group (type optparse.OptionGroup) of Hachoir
    library options.
    """

    common = OptionGroup(parser, "Hachoir library", \
        "Configure Hachoir library")
    common.add_option("--verbose", help="Verbose mode",
        default=False, action="store_true")
    common.add_option("--log", help="Write log in a file",
        type="string", action="callback", callback=_setLogFilename)
    common.add_option("--quiet", help="Quiet mode (don't display warning)",
        default=False, action="store_true")
    common.add_option("--debug", help="Debug mode",
        default=False, action="store_true")
    if console_ui:
        common.add_option("--parser-list", help="Display parser list",
            action="callback", callback=displayParserList)
    return common

def initHachoir(option=None):
    # Setup locales
    try:
        locale.setlocale(locale.LC_ALL, "")
    except (locale.Error, IOError):
        warning("Fail to setup locales.")

    # Configure Hachoir using "option"
    if option:
        if option.quiet:
          config.quiet = True
        if option.verbose:
          config.verbose = True
        if option.debug:
          config.debug = True

def createParser(filename, mime=None):
    # Load parsers (do nothing if function has already been call)
    loadParserPlugins()

    # Create input stream
    stream = FileInputStream(filename)

    # Create field set
    field_set = guessParser(stream, filename, mime=mime)
    return field_set

def runBenchmark(count, func, *args, **kw):
    stat = []
    for index in xrange(count):
        before = time.time()
        func(*args, **kw)
        after = time.time()
        stat.append(after-before)
    geometric = math.exp(sum(sorted(map(math.log,stat))) / count)
    print "Benchmark:"
    print " - best:    %s" % humanDurationNanosec(min(stat) * 1000000000)
    print " - average: %s (geometric mean)" % humanDurationNanosec(geometric * 1000000000)
    print " - worst:   %s" % humanDurationNanosec(max(stat) * 1000000000)
    print " - total:   %s" % humanDurationNanosec(sum(stat) * 1000000000)

