from xml.dom.minidom import getDOMImplementation
from xml.dom.ext import PrettyPrint
from hachoir.field import (Enum, GenericInteger,
    Int8, Int16, Int32, Int64,
    UInt8, UInt16, UInt32, UInt64,
    PaddingBits,PaddingBytes,
    String, CString, PascalString32,
    RawBytes,
    RawBits, Bit, Bits)
import hachoir
from hachoir.endian import BIG_ENDIAN
from hachoir.bits import long2bin
from hachoir.error import error
import base64
import os
import re

def raw2xml(content):
    special_chars = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        '"': "&quote;",
        "'": "&apos;",
    }
    output = ""
    escapped = 0
    for character in content:
        if 32 <= ord(character) <= 127:
            if character in special_chars:
                output += special_chars[character]
            else:
                output += character
        else:
            output += "&#"+str(ord(character))+";"
            escapped += 1
        if 10 < escapped:
            content = base64.encodestring(content).strip()
            return (True, content)
    return (False, output)

class ExportXML:
    upper_regex = re.compile("([A-Z])")

    def __init__(self):
        self.document = None
        self.class_handler = {
            Enum: self.writeEnum,
            PaddingBits: self.writePadding,
            PaddingBytes: self.writePadding,
            String: self.writeString,
            CString: self.writeString,
            PascalString32: self.writeString,
            GenericInteger: self.writeInteger,
            Int8: self.writeInteger,
            Int16: self.writeInteger,
            Int32: self.writeInteger,
            Int64: self.writeInteger,
            UInt8: self.writeInteger,
            UInt16: self.writeInteger,
            UInt32: self.writeInteger,
            UInt64: self.writeInteger,
            RawBits: self.writeBits,
            Bit: self.writeBit,
            Bits: self.writeBits,
            RawBytes:  self.writeRawBytes
        }
        self.write_data = True
        self.enum_uniq_id = {}

    def writeCommonAttributs(self, field, node):
        node.setAttribute("name", field.name)
        node.setAttribute("size", str(field.size))
        node.setAttribute("address", str(field.address))
        if field.description is not None:
            node.setAttribute("description", field.description)

    def writeFieldSet(self, field):
        node = self.document.createElement("field_set")
        node.setAttribute("module", field.__class__.__module__)
        node.setAttribute("type", field.__class__.__name__)
        return node

    def writeDefault(self, field):
        # Use class name
        name = field.__class__.__name__

        # Convert "NameWithUpper" to "Name_With_Upper", and then convert
        # to lower case: "name_with_upper"
        name = name[0] + self.upper_regex.sub(r'_\1', name[1:])
        name = name.lower()
        return self.document.createElement(name)

    def writeInteger(self, field):
        if field.__class__ is Enum:
            node = self.document.createElement("enum")
        else:
            node = self.document.createElement("integer")
        node.setAttribute("signed", str(field.signed))
        if self.write_data:
            data = self.document.createTextNode(str(field.value))
            node.appendChild(data)
        return node

    def writeEnumDict(self, enum):
        parent = self.document.createElement("enum_dict")
        uniq_id = id(enum)
        if uniq_id in self.enum_uniq_id:
            uniq_id = self.enum_uniq_id[uniq_id]
        else:
            uniq_id = str(1+len(self.enum_uniq_id))
            self.enum_uniq_id[uniq_id] = uniq_id
        parent.setAttribute("id", uniq_id)
        for key, value in enum.items():
            node = self.document.createElement("key")
            node.setAttribute("value", str(key))
            data = self.document.createTextNode(str(value))
            node.appendChild(data)
            parent.appendChild(node)
        return (uniq_id, parent)

    def writeEnum(self, field):
        node = self.writeInteger(field)
        if self.write_data:
            enum = field.getEnum()
            uniq_id, data = self.writeEnumDict(enum)
            node.setAttribute("enum", uniq_id)
            self.document.documentElement.appendChild(data)
        return node

    def writePadding(self, field):
        node = self.document.createElement("padding")
#        if field.pattern:
#            node.setAttribute("pattern", field.pattern)
        return node

    def writeString(self, field):
        node = self.document.createElement("string")
        node.setAttribute("format", field.format)
        node.setAttribute("length", str(field.length))
        if field.charset is not None:
            node.setAttribute("charset", field.charset)
#        if field.pattern:
#            node.setAttribute("pattern", field.pattern)
        return node

    def writeBit(self, field):
        if field.__class__ == Bit:
            node = self.document.createElement("bit")
        else:
            node = self.document.createElement("raw_bit")

        node = self.document.createElement("bit")
        if self.write_data:
            if field.value:
                content = "1"
            else:
                content = "0"
            data = self.document.createTextNode(content)
            node.appendChild(data)
        return node

    def writeBits(self, field):
        if field.__class__ == Bits:
            node = self.document.createElement("bits")
        else:
            node = self.document.createElement("raw_bits")
        node.setAttribute("bits", str(field.size))
        if self.write_data:
            content = long2bin(field.size, field.value, BIG_ENDIAN)
            data = self.document.createTextNode(content)
            node.appendChild(data)
        return node

    def writeRawBytes(self, field):
        node = self.document.createElement("bytes")
        if self.write_data:
            if field.size < 40*8:
                use_base64, content = raw2xml(field.value)
                if use_base64:
                    node.setAttribute("base64", "1")
            else:
                node.setAttribute("truncated", "1")
                content = "***DATA TOO BIG***"
            data = self.document.createTextNode(content)
            node.appendChild(data)
        return node

    def writeNodes(self, parentparent, field_set, max_depth, depth=0):
        parent = self.document.createElement("field_set")
        parent.setAttribute("module", field_set.__class__.__module__)
        parent.setAttribute("type", field_set.__class__.__name__)
        self.writeCommonAttributs(field_set, parent)
        parentparent.appendChild(parent)


        for field in field_set:
            if field.is_field_set:
                self.writeNodes(parent, field, max_depth, depth+1)
            elif field.size:
                cls = field.__class__
                if cls in self.class_handler:
                    handler = self.class_handler[cls]
                else:
                    handler = self.writeDefault
                node = handler(field)
                self.writeCommonAttributs(field, node)
                parent.appendChild(node)

    def __call__(self, filename, field_set, max_depth):
        filename = os.path.expanduser(filename)
        if os.access(filename, os.F_OK):
            error("File already exists: %s" % filename)
            return False
        output = open(filename, "w")
        impl = getDOMImplementation()
        self.document = impl.createDocument(None, "hachoir", None)
        root = self.document.documentElement
        root.setAttribute("version", hachoir.__version__)
        self.writeNodes(root, field_set, max_depth)
        PrettyPrint(self.document, output)
        output.close()
        return True

