"""
Parent of all (field) classes in Hachoir: Field.
"""

from hachoir.error import HachoirError

class FieldError(HachoirError):
    """
    Error raised by a L{Field}.

    @see: L{HachoirError}
    """
    pass

class MissingField(KeyError, FieldError):
    def __init__(self, field, key):
        KeyError.__init__(self)
        self.field = field
        self.key = key

    def __str__(self):
            return "Can't get %s['%s']" % (self.field.path, self.key)

class Field(object):
    # static size can have two differents value: None (no static size), an
    # integer (number of bits), or a function which returns an integer.
    #
    # This function receives exactly the same arguments than the constructor
    # except the first one (one). Example of function:
    #    static_size = staticmethod(lambda *args, **kw: args[1])
    static_size = None

    # Indicate if this field contains other fields (is a field set) or not
    is_field_set = False

    def __init__(self, parent, name, size=None, description=None):
        """
        Set default class attributes, set right address if None address is
        given.

        @param parent: Parent field of this field
        @type parent: L{Field}|None
        @param name: Name of the field, have to be unique in parent. If it ends
            with "[]", end will be replaced with "[new_id]" (eg. "raw[]"
            becomes "raw[0]", next will be "raw[1]", and then "raw[2]", etc.)
        @type name: str
        @param size: Size of the field in bit (can be None, so it
            will be computed later)
        @type size: int|None
        @param address: Address in bit relative to the parent absolute address
        @type address: int|None
        @param description: Optional string description
        @type description: str|None
        """
        assert (size is None) or (0 <= size)
        self._parent = parent
        self._name = name
        self._value = None
        if parent:
            assert issubclass(parent.__class__, Field)
            self._address = parent._current_size
        else:
            self._address = 0
        self._size = size
        self._description = description

    def _createDescription(self):
        return ""
    def _getDescription(self):
        if self._description is None:
            self._description = self._createDescription()
        return self._description
    description = property(_getDescription,
    doc="Description of the field (string)")

    def __str__(self):
        return self.display

    def hasValue(self):
        return hasattr(self, "_createValue")
    def _getValue(self):
        if self._value is None:
            self._value = self._createValue()
        return self._value
    def _setValue(self, new_value):
        self._parent.raiseEvent("set-field-value", self, new_value)
    value = property(_getValue, _setValue, "Value of field")

    def _getParent(self):
        return self._parent
    parent = property(_getParent,
    doc="Parent (a field set) of this field")

    def _getDisplay(self):
        return str(self.value)
    display = property(lambda self: self._getDisplay(),
    doc="Short string which represents field content")

    def _getName(self):
        return self._name
    name = property(_getName,
    doc="Field name (unique in its parent field set list)")

    def _getIndex(self):
        if not self._parent:
            return None
        return self._parent._fields.indexOf(self._name)
    index = property(_getIndex)

    def _getPath(self):
        names = [ '' ]
        def rec(field):
            if field._parent:
                rec(field._parent)
                names.append(field.name)
        rec(self)
        return '/'.join(names)
    path = property(_getPath,
    doc="Full path of the field starting at root field")

    def _getAddress(self):
        return self._address
    address = property(_getAddress,
    doc="Relative address in bit to parent address")

    def _getAbsoluteAddress(self):
        address = self._address
        current = self._parent
        while current:
            address += current._address
            current = current._parent
        return address
    absolute_address = property(_getAbsoluteAddress,
    doc="Absolute address (from stream beginning) in bit")

    def _getSize(self):
        return self._size
    size = property(_getSize, doc="Content size in bit")

    def _getField(self, name, const):
        if name.strip("."):
            return None
        field = self
        for index in xrange(1, len(name)):
            field = field._parent
            if field is None:
                break
        return field

    def getField(self, key, const=True):
        if key:
            if key[0] == "/":
                if self._parent:
                    current = self._parent.root
                else:
                    current = self
                if len(key) == 1:
                    return current
                key = key[1:]
            else:
                current = self
            for part in key.split("/"):
                field = current._getField(part, const)
                if field is None:
                    raise MissingField(current, part)
                current = field
            return current
        raise KeyError("Key must not be an empty string!")

    def __getitem__(self, key):
        return self.getField(key, False)

    def __contains__(self, name):
        try:
            self[name]
            return True
        except FieldError:
            return False
