from hachoir.field import (Field, FieldError,
    createRawField, createPaddingField)
from hachoir.event_handler import EventHandler
from hachoir.dict import Dict
from hachoir.endian import BIG_ENDIAN, LITTLE_ENDIAN
from hachoir.stream import InputStream
from hachoir.error import HachoirError, error, warning, info
from hachoir.tools import lower_bound
import hachoir.config as config

class ParserError(FieldError):
    """
    Error raised by a L{GenericFieldSet} (or L{FieldSet}/L{Parser}).

    @see: L{FieldError}
    """
    pass

class MatchError(FieldError):
    """
    Error raised by a L{FieldSet} or a L{Parser} when the stream content
    doesn't match to file format.

    @see: L{FieldError}
    """
    pass

class GenericFieldSet(Field):
    """
    Ordered list of fields. Use operator [] to access fields using their
    name (field names are unique in a field set, but not in the whole
    document).

    Class attributes:
    - endian: Bytes order (L{BIG_ENDIAN} or L{LITTLE_ENDIAN}). Optional if the
      field set has a parent ;
    - static_size: (optional) Size of FieldSet in bits. This attribute should
      be used in parser of constant size.

    Instance attributes/methods:
    - _fields: Ordered dictionnary of all fields, may be incomplete
      because feeded when a field is requested ;
    - stream: Input stream used to feed fields' value
    - root: The root of all field sets ;
    - __len__(): Number of fields, may need to create field set ;
    - __getitem(): Get an field by it's name or it's path.

    And attributes inherited from Field class:
    - parent: Parent field (may be None if it's the root) ;
    - name: Field name (unique in parent field set) ;
    - value: The field set ;
    - address: Field address (in bits) relative to parent ;
    - description: A string describing the content (can be None) ;
    - size: Size of field set in bits, may need to create field set.

    Event handling:
    - "connectEvent": Connect an handler to an event ;
    - "raiseEvent": Raise an event.

    To implement a new field set, you need to:
    - create a class which inherite from FieldSet ;
    - write createFields() method using lines like:
         yield Class(self, "name", ...) ;
    - and maybe set endian and static_size class attributes.
    """

    is_field_set = True
    endian = None

    def __init__(self, parent, name, stream, description=None, size=None):
        """
        Constructor
        @param parent: Parent field set, None for root parser
        @param name: Name of the field, have to be unique in parent. If it ends
            with "[]", end will be replaced with "[new_id]" (eg. "raw[]"
            becomes "raw[0]", next will be "raw[1]", and then "raw[2]", etc.)
        @type name: str
        @param stream: Input stream from which data are read
        @type stream: L{InputStream}
        @param description: Optional string description
        @type description: str|None
        @param size: Size in bits. If it's None, size will be computed. You
            can also set size with class attribute static_size
        """

        # Set field set size
        if self.static_size is not None:
            assert isinstance(self.static_size, (int, long))
            size = self.static_size

        # Make some tests on arguments
        assert not parent or issubclass(parent.__class__, GenericFieldSet)
        assert issubclass(stream.__class__, InputStream)

        # Call parent class constructor
        Field.__init__(self, parent, name, size=size, description=description)

        # Set endian
        if self.endian is None:
            assert parent is not None and parent.endian is not None
            self.endian = parent.endian
        assert self.endian in (BIG_ENDIAN, LITTLE_ENDIAN)

        # Create field list dictionnary
        self._fields = Dict()

        self._event_handler = None
        self._field_generator = self.createFields()
        self._field_array_count = {}
        self._current_size = 0
        if parent:
            # This field set is one of the root leaf
            self.root = parent.root
            self.stream = parent.stream
        else:
            # This field set is the root of the tree
            self.root = self
            assert issubclass(stream.__class__, InputStream)
            self.stream = stream
            self._global_event_handler = None

    def connectEvent(self, event_name, handler, local=True):
        assert event_name in (\
            # Callback prototype: def f(field)
            # Called when new value is already set
            "field-value-changed",

            # Callback prototype: def f(field)
            # Called when field size is already set
            "field-resized",

            # A new field has been insered in the field set
            # Callback prototype: def f(index, new_field)
            "field-insered",

            # Callback prototype: def f(old_field, new_field)
            # Called when new field is already in field set
            "field-replaced",

            # Callback prototype: def f(field, new_value)
            # Called to ask to set new value
            "set-field-value"
        )
        if local:
            if self._event_handler is None:
                self._event_handler = EventHandler()
            self._event_handler.connect(event_name, handler)
        else:
            if self.root._global_event_handler is None:
                self.root._global_event_handler = EventHandler()
            self.root._global_event_handler.connect(event_name, handler)

    def reset(self):
        """
        Reset a field set:
         * clear fields ;
         * restart field generator ;
         * set current size to zero ;
         * clear field array count.

        But keep: name, value, description and size.
        """
        self._fields = Dict()
        self._field_generator = self.createFields()
        self._field_array_count = {}
        self._current_size = 0

    def __str__(self):
        return '%s <path=%s, current_size=%s, current length=%s>' % \
            (self.__class__.__name__, self.path, self._current_size, len(self._fields))

    def raiseEvent(self, event_name, *args):
        # Transfer event to local listeners
        if self._event_handler is not None:
            self._event_handler.raiseEvent(event_name, *args)

        # Transfer event to global listeners
        if self.root._global_event_handler is not None:
            self.root._global_event_handler.raiseEvent(event_name, *args)

    def __nonzero__(self):
        """
        Method called by code like "if field_set: (...)".
        Always returns True
        """
        return True


    def __len__(self):
        """
        Returns number of fields, may need to create all fields
        if it's not done yet.
        """
        if self._field_generator is not None:
            self._feedAll()
        return len(self._fields)

    def _getCurrentLength(self):
        return len(self._fields)
    current_length = property(_getCurrentLength)

    def _getSize(self):
        if self._size is None:
            self._feedAll()
        return self._size
    size = property(_getSize, doc="Size in bit, may create all fields to get size")

    def _getCurrentSize(self):
        assert not(self.done)
        return self._current_size
    current_size = property(_getCurrentSize)

    def _setUniqFieldName(self, field):
        if field._name.endswith("[]"):
            name = field._name[:-2]
            if name in self._field_array_count:
                self._field_array_count[name] += 1
            else:
                self._field_array_count[name] = 0
            field._name = name + "[%u]" % self._field_array_count[name]

    def _addField(self, field):
        # Replace "name[]" with "name[<index>]"
        self._setUniqFieldName(field)

        # Set field address
        field._address = self._current_size

#        if config.debug:
#            addr = field.absolute_address
#            info("* Instanciate %s (addr=%s.%s, size=%s bits)" \
#                % (field.path, addr/8, addr%8, field.size))

        # WARNING: Reading "field.size" here may takes long time
        # since it may create fields in field to compute total size

        # Check field size
        if self.stream.size < self._current_size + field.size:
            size = field.size
            addr = field.absolute_address
            raise ParserError("Field %s is too big (%s.%s bytes at %s.%s)!" % \
                (field.path, size/8, size%8, addr/8, addr%8))

        # Update total field size
        self._current_size += field.size

        if config.debug:
            addr = field.absolute_address
            size = field._size
            if field._description:
                desc = ": "+field._description
            else:
                desc = ""
            info("[+] %s%s (addr=%s.%s, size=%s.%s)" \
                % (field.path, desc, addr/8, addr%8, size/8, size%8))

        # Append field to the field list
        self._fields.append(field._name, field)

    def _feed(self):
#        if config.debug:
#            info("* %s.feed()" % self.path)
        field = self._field_generator.next()
        self._addField(field)
        return field

    def _getField(self, name, const):
        field = Field._getField(self, name, const)
        if field is None:
            if name in self._fields:
                field = self._fields[name]
            elif self._field_generator is not None and not const:
                field = self._feedUntil(name)
        return field

    def getField(self, key, const=True):
        if isinstance(key, (int, long)):
            if key < 0:
                raise KeyError("Key must be positive!")
            if not const:
                self.readFirstFields(key+1)
            return self._fields.values[key]
        return Field.getField(self, key, const)

    def _stopFeeding(self):
        self._field_generator = None

        # Parser is smaller than than file?
        if self.root == self and self._current_size < self._size:
            # Add a raw field
            size = self._size-self._current_size
            raw = createRawField(self, size, "raw[]", "End of file")
            self._addField(raw)
            assert self._current_size == self._size

            # And display a warning
            warning("File is bigger than the parser (%u.%u bytes at the end)!"
                % (size/8, size%8))
        elif (self._size is not None) and (self._size != self._current_size):
            raise ParserError(
                "Invalid size for field set %s: %s instead of %s!"
                % (self.path, self._current_size, self._size))
        self._size = self._current_size

    def _parserError(self, exception):
        error("Parser error: stop parser at \"%s\". Error: %s" \
            % (self.path, exception))
        if self._size is None:
            raise ParserError("Unable to fix parser error (unable to guess parser size).")
        size = self._size - self._current_size
        if size < 0:
            raise ParserError("Unable to fix parser error (invalid size: %s)." % size)
        if size == 0:
            return None
        field = createRawField(self, size, "raw[]")
        self._addField(field)
        self._stopFeeding()
        return field

    def _feedUntil(self, field_name):
        """
        Return the field if it was found, None else
        """
        try:
            while True:
                field = self._feed()
                if field.name == field_name:
                    return field
        except FieldError, err:
            self._parserError(err)
        except StopIteration:
            self._stopFeeding()
        return None

    def readFirstFields(self, number):
        """
        Read first number fields if they are not read yet.
        """
        if self._field_generator is None:
            return
        number = number - len(self._fields)
        if 0 < number:
            return self.readMoreFields(number)
        else:
            return 0

    def readMoreFields(self, number):
        """
        Read more number fields, or do nothing if parsing is done.
        """
        if self._field_generator is None:
            return
        added = 0
        try:
            for index in xrange(number):
                self._feed()
                added += 1
        except FieldError, err:
            self._parserError(err)
            added += 1
        except StopIteration:
            self._stopFeeding()
        return added

    def _feedAll(self):
        if self._field_generator is None:
            return
        try:
            while True:
                self._feed()
        except FieldError, err:
            self._parserError(err)
        except StopIteration:
            self._stopFeeding()

    def __iter__(self):
        """
        Create a generator to iterate on each field, may create new
        fields when needed
        """
        try:
            done = 0
            while True:
                if done == len(self._fields):
                    if self._field_generator is None:
                        break
                    self._feed()
                for field in self._fields.values[done:]:
                    yield field
                    done += 1
        except FieldError, err:
            field = self._parserError(err)
            if field:
                yield field
            return
        except StopIteration:
            self._stopFeeding()

    def createFields(self):
        """
        DON'T CALL THIS FUNCTION DIRECTLY! Use: L{__iter__()} or
        L{__getitem__()} to access fields.

        This function have to be implemented in concrete field set.
        """
        raise NotImplementedError

    def _isDone(self):
        return (self._field_generator is None)
    done = property(_isDone, doc="Boolean to know if parsing is done or not")


    #
    # FieldSet_SeekUtility
    #
    def seekBit(self, address, name="padding[]",
    description=None, relative=True):
        """
        Create a field to seek to specified address,
        or None if it's not needed.

        May raise an (ParserError) exception if address is invalid.
        """
        if relative:
            nbits = address - self._current_size
        else:
            nbits = address - (self.absolute_address + self._current_size)
        if nbits < 0:
            raise ParserError("Seek error, unable to go back!")
        if 0 < nbits:
            return createPaddingField(self, nbits, name, description)
        else:
            return None

    def seekByte(self, address, name="padding[]", description=None, relative=True):
        """
        Same as seekBit(), but with address in byte.
        """
        return self.seekBit(address * 8, name, description, relative)

    #
    # RandomAccessFieldSet
    #
    def replaceField(self, name, new_fields):
        # TODO: Check in self and not self.field
        # Problem is that "generator is already executing"
        if name not in self._fields:
            raise ParserError("Unable to replace %s: field doesn't exist!" % name)
        assert 1 <= len(new_fields)
        old_field = self[name]
        total_size = sum( (field.size for field in new_fields) )
        if old_field.size != total_size:
            raise ParserError( \
                "Unable to replace %s: " \
                "new field(s) hasn't same size (%u bits instead of %u bits)!" \
                % (name, total_size, old_field.size))
        field = new_fields[0]
        self._setUniqFieldName(field)
        field._address = old_field.address
        if field.name != name and field.name in self._fields:
            raise ParserError( \
                "Unable to replace %s: name \"%s\" is already used!" \
                % (name, field.name))
        self._fields.replace(name, field.name, field)
        self.raiseEvent("field-replaced", old_field, field)
        if 1 < len(new_fields):
            index = self._fields.indexOf(new_fields[0].name)+1
            address = field.address + field.size
            for field in new_fields[1:]:
                self._setUniqFieldName(field)
                field._address = address
                if field.name in self._fields:
                    raise ParserError( \
                        "Unable to replace %s: name \"%s\" is already used!" \
                        % (name, field.name))
                self._fields.insert(index, field.name, field)
                self.raiseEvent("field-insered", index, field)
                index += 1
                address += field.size

    def getExistingFieldByAddress(self, address):
        """
        Only search in existing fields
        """
        if address < self._current_size:
            i = lower_bound(self._fields.values, lambda x: x.address + x.size <= address)
            if i is not None:
                return self._fields.values[i]
        return None

    def writeFieldsIn(self, old_field, address, new_fields):
        """
        Can only write in existing fields (address < self._current_size)
        """

        # Check size
        total_size = sum( (field.size for field in new_fields) )
        if old_field.size < total_size:
            raise ParserError( \
                "Unable to write fields at address %s " \
                "(too big)!" % (address))

        # Need padding before?
        replace = []
        size = address - old_field.address
        assert 0 <= size
        if 0 < size:
            padding = createPaddingField(self, size)
            padding._address = old_field.address
            replace.append(padding)

        # Set fields address
        for field in new_fields:
            field._address = address
            address += field.size
            replace.append(field)

        # Need padding after?
        size = (old_field.address + old_field.size) - address
        assert 0 <= size
        if 0 < size:
            padding = createPaddingField(self, size)
            padding._address = address
            replace.append(padding)

        self.replaceField(old_field.name, replace)

