from hachoir.field import GenericFieldSet, ParserError
from hachoir.endian import BIG_ENDIAN, LITTLE_ENDIAN

class Parser(GenericFieldSet):
    """
    A parser is the root of all other fields. It create first level of fields
    and have special attributes and methods:
    - tags: dictionnary with keys:
      - "file_ext": classical file extensions (string or tuple of strings) ;
      - "mime": MIME type(s) (string or tuple of strings) ;
      - "description": String describing the parser.
    - endian: Byte order (L{BIG_ENDIAN} or L{LITTLE_ENDIAN}) of input data ;
    - stream: Data input stream (set in L{__init__()}).

    Default values:
    - size: Field set size will be size of input stream ;
    - mime_type: First MIME type of tags["mime"] (if it does exist,
      None otherwise).
    """

    def __init__(self, stream):
        """
        Parser constructor

        @param stream: Data input stream (see L{InputStream})
        @param description: (optional) String description
        """
        # Check arguments
        assert hasattr(self, "endian") \
            and self.endian in (BIG_ENDIAN, LITTLE_ENDIAN)
        assert hasattr(self, "tags")
        assert "description" in self.tags and self.tags["description"] != ""
        assert "min_size" in self.tags and 0 < self.tags["min_size"]

        # Call parent constructor
        GenericFieldSet.__init__(self, None, "root", stream,
            self.tags["description"], size=stream.size)

        # Set (default) MIME type
        if "mime" in self.tags:
            assert (isinstance(self.tags["mime"], tuple)
                 or isinstance(self.tags["mime"], list))
            self._mime_type = self.tags["mime"][0]
        else:
            self._mime_type = None

    def validate(self):
        """
        Check that the parser is able to parse the stream. Valid results:
        - True: stream looks valid ;
        - False: stream is invalid ;
        - str: string describing the error.
        """
        raise NotImplementedError()

    def _getPath(self):
        return "/"
    path = property(_getPath)

    def _getMimeType(self):
        return self._mime_type
    mime_type = property(_getMimeType)

