"""
String field classes:
- String: Fixed length string (no prefix/no suffix) ;
- CString: String which ends with nul byte ("\0") ;
- UnixLine: Unix line of text, string which ends with "\n" ;
- PascalString8, PascalString16, PascalString32: String prefixed with
  length written in a 8, 16, 32-bit integer (use parent endian).

Constructor has optional arguments:
- strip: value can be a string or True ;
- charset: if set, convert string to unicode using this charset (in "replace"
  mode which replace all buggy characters with ".").

Note: For PascalStringXX, prefixed value is the number of bytes and not
      of characters!
"""

from hachoir.field import Field, FieldError, Bytes
from hachoir.error import warning
from hachoir import config

class GenericString(Bytes):
    """ Generic string class """
    __pascal_format_to_size = {
        "Pascal8": 8,
        "Pascal16": 16,
        "Pascal32": 32
    }

    def __init__(self, parent, name,
    format, description=None, strip=None, charset=None, length=None):
        Bytes.__init__(self, parent, name, 1, description)

        # Is format valid?
        assert format in ("C", "UnixLine",
            "fixed", "Pascal8", "Pascal16", "Pascal32")

        # Arbitrary limit, just to catch some bugs...
        assert (format != "fixed") or (1 <= length <= 0xffff)

        # Forbid charset "UTF-16": use UTF-16LE or UTF-16BE
        assert charset is None or charset.upper() not in ("UTF-16", "UTF16")

        # Save attributes
        self._format = format
        self._charset = charset
        self._strip = strip
        if self._charset:
            self._is_utf16 = self._charset.upper().startswith("UTF-16")
        else:
            self._is_utf16 = False

        # Set begin/end offset (in bytes)
        self._begin_offset = 0
        self._end_offset = 0
        if self._format in ("C", "UnixLine"):
            self._end_offset = 1
        elif self._format in self.__pascal_format_to_size:
            self._begin_offset = self.__pascal_format_to_size[self._format]/8

        if length is None:
            self._computeSize()
        else:
            self._length = length
            if self._is_utf16:
                self._size = self._length * 16
            else:
                self._size = self._length * 8
            self._size += (self._begin_offset + self._end_offset) * 8
        assert 0 <= self._length

    def _computeSize(self):
        if self._format == "C":
            # C: Search nul byte ("\0")
            start = self.absolute_address
            stop = self._parent.stream.searchBytes("\0", start)
            if stop is None:
                raise FieldError(
                "Can't find end of C string %s ('\\0')!" % self.path)
            self._size = stop - start + 8
        elif self._format == "UnixLine":
            # Unix: Search new line character ("\n")
            start = self.absolute_address
            stop = self._parent.stream.searchBytes("\n", start)
            if stop is None:
                raise FieldError(
                "Can't find end of Unix line %s ('\\n')!" % self.path)
            self._size = stop - start + 8
        else:
            # Pascal: Read length in bytes
            assert self._format in self.__pascal_format_to_size
            endian = self._parent.endian
            size = self.__pascal_format_to_size[self._format]
            value = self._parent.stream.readBits(
                self.absolute_address, size, endian)
            self._size = (self._begin_offset + value) * 8
            if self._is_utf16:
                self._length = value // 2
            else:
                self._length = value

        # Save field length
        self._length = (self._size / 8) \
            - self._begin_offset - self._end_offset

    def _getFormat(self):
        return self._format
    format = property(_getFormat, doc="String format (eg. 'C')")

    def _getCharset(self):
        return self._charset
    charset = property(_getCharset, doc="String charset (eg. 'ISO-8859-1')")

    def _getLength(self):
        return self._length
    length = property(_getLength, doc="String length in characters")

    def _createValue(self):
        # Compute size in bytes
        size = self._size/8 - (self._begin_offset + self._end_offset)
        if size == 0:
            # Empty string
            if self._charset:
                return u""
            else:
                return ""

        # Read bytes
        text = self._parent.stream.readBytes( \
            self.absolute_address + self._begin_offset*8, size)

        # Convert to unicode if needed
        if self._charset:
            try:
                text = unicode(text, self._charset)
            except UnicodeDecodeError, err:
                # On error, stay in str type
                warning("Unable to convert string %s to Unicode: %s"
                    % (self.path, err))
                self._charset = None

        # Strip string if needed
        if self._strip:
            if isinstance(self._strip, str):
                text = text.strip(self._strip)
            else:
                text = text.strip()
        return text

    def _getDisplay(self):
        value = self.value
        if config.max_string_length < len(value):
            value = value[:config.max_string_length] + "(...)"
        return '"%s"' % value

def StringFactory(format):
    class NewString(GenericString):
        def __init__(self, parent, name, description=None,
            strip=None,
            charset=None):
            GenericString.__init__(self, parent, name,
                format, description, strip, charset)
    return NewString

# String which ends with nul byte ("\0")
CString = StringFactory("C")

# Unix line of text: string which ends with "\n" (ASCII 0x0A)
UnixLine = StringFactory("UnixLine")

# String prefixed with length written in a 8-bit integer
PascalString8 = StringFactory("Pascal8")

# String prefixed with length written in a 16-bit integer (use parent endian)
PascalString16 = StringFactory("Pascal16")

# String prefixed with length written in a 32-bit integer (use parent endian)
PascalString32 = StringFactory("Pascal32")

class String(GenericString):
    """ String with fixed length (length in bytes) """
    static_size = staticmethod(lambda *args, **kw: args[1]*8)

    def __init__(self, parent, name, length, description=None,
    strip=None, charset=None):
        GenericString.__init__(self, parent, name,
            "fixed", description, strip, charset, length)

