from hachoir.metadata.metadata import (Metadata, MultipleMetadata,
    registerExtractor)
from hachoir.parser.archive import (Bzip2File, GzipFile, TarFile, ZipFile)
from hachoir.tools import humanFilesize, humanUnixAttributes

class FileMetadata(Metadata):
    header = "File"
    def __init__(self):
        Metadata.__init__(self)
        self.register("filename", 300, "File name")
        self.register("file_size", 301, "File size", handler=humanFilesize)
        self.register("file_attr", 400, "File attributes")
        self.register("file_type", 401, "File type")

class CompressedFileMetadata(FileMetadata):
    header = "Compressed file"

    def __init__(self, **kw):
        FileMetadata.__init__(self, **kw)
        self.register("compr_size", 310, "Compressed file size", handler=humanFilesize)
        self.register("compr_rate", 320, "Compression rate")

    def postExtract(self):
        if hasattr(self, "file_size") and hasattr(self, "compr_size"):
            rate = 100 - float(self.compr_size) * 100 / self.file_size
            self.compr_rate = "%.1f%%" % rate

class Bzip2Metadata(CompressedFileMetadata):
    def extract(self, zip):
        self.compr_size = zip["content"].size/8

class GzipMetadata(CompressedFileMetadata):
    def extract(self, gzip):
        if "filename" in gzip:
            self.filename = gzip["filename"].value
        self.compr_size = gzip["data"].size/8
        self.compression = gzip["compression"].display
        self.file_size = gzip["size"].value
        self.last_modification = gzip["mtime"].display
        self.producer = "Created on %s" % gzip["os"].display
        if "comment" in gzip:
            self.comment = gzip["comment"].value
        self.postExtract()

class ZipMetadata(MultipleMetadata):
    def extract(self, zip):
        for field in zip:
            if not field.name.startswith("file["):
                continue
            meta = CompressedFileMetadata()
            meta.filename = field["filename"].value
            meta.file_size = field["uncompressed_size"].value
            meta.creation_date = field["last_mod"].display
            meta.compression = field["compression"].display
            self.addGroup(field.name, meta, "File \"%s\"" % meta.filename)

class TarMetadata(MultipleMetadata):
    def extract(self, tar):
        for field in tar:
            if field.isEmpty():
                return
            meta = FileMetadata()
            meta.filename = field["name"].value
            meta.file_size = field.getOctal("size")
            meta.creation_date = field.getDatetime()
            meta.file_attr = humanUnixAttributes(field.getOctal("mode"))
            meta.file_type = field["type"].display
            meta.author = "%s (uid=%s), group %s (gid=%s)" %\
                (field["uname"].value, field.getOctal("uid"),
                 field["gname"].value, field.getOctal("gid"))
            self.addGroup(field.name, meta, "File \"%s\"" % meta.filename)

registerExtractor(GzipFile, GzipMetadata)
registerExtractor(Bzip2File, Bzip2Metadata)
registerExtractor(TarFile, TarMetadata)
registerExtractor(ZipFile, ZipMetadata)

