from hachoir.metadata.metadata import Metadata, registerExtractor
from hachoir.parser.audio import (AuFile, Mp3File)
from hachoir.tools import humanFrequency, humanDuration, humanBitSize

class AudioMetadata(Metadata):
    header = "Audio"
    nb_channel_name = {
        1: "mono",
        2: "stereo"
    }

    def __init__(self):
        Metadata.__init__(self)
        self.register("album", 200, "Album")
        self.register("track_number", 201, "Track number")

        self.register("nb_channel", 300, "Channel", handler=self.channelHandler)
        self.register("sample_rate", 301, "Sample rate", handler=humanFrequency)

        self.register("bits_per_sec", 601, "Bits/second", handler=humanBitSize)

    def channelHandler(self, value):
        if value in self.nb_channel_name:
            return self.nb_channel_name[value]
        else:
            return value

class AuMetadata(AudioMetadata):
    def extract(self, audio):
        self.sample_rate = audio["sample_rate"].value
        self.nb_channel = audio["channels"].value
        self.compression = audio["codec"].display
        if "info" in audio:
            self.comment = audio["info"].value
        bit_rate = audio.getBitRate()
        if bit_rate:
            self.bits_per_sec = bit_rate * self.nb_channel * self.sample_rate
            self.duration = audio["audio_data"].size * 1000 / self.bits_per_sec

class Mp3Metadata(AudioMetadata):
    tag_to_key = {
        # ID3 version 2.2
        "TP1": ("text", "author"),
        "COM": ("text", "comment"),
        "TEN": ("text", "producer"),
        "TRK": ("text", "track_number"),
        "TAL": ("text", "album"),
        "TI2": ("text", "title"),

        # ID3 version 2.3+
        "TPE1": ("text", "author"),
        "COMM": ("text", "comment"),
        "TENC": ("text", "producer"),
        "TRCK": ("text", "track_number"),
        "TALB": ("text", "album"),
        "TIT2": ("text", "title"),
        "WXXX": ("text", "url")
    }

    def processID3v2(self, field):
        # Tag is interresting?
        tag = field["tag"].value
        if tag not in self.tag_to_key:
            return

        field_type, key = self.tag_to_key[tag]

        # Text is empty?
        if "text" not in field["content"]:
            return

        # Save value
        value = field["content/text"].value
        setattr(self, key, value)

    def readID3v2(self, id3):
        for field in id3:
            if field.is_field_set and "tag" in field:
                self.processID3v2(field)

    def extract(self, mp3):
        self.sample_rate = mp3.getSampleRate()
        self.nb_channel = mp3["channel_mode"].display
        self.compression = "MPEG version %s layer %s" % \
            (mp3["version"].display, mp3["layer"].display)
        self.bits_per_sec = mp3.getKBitRate()*1000
        if "id3v1" in mp3:
            id3 = mp3["id3v1"]
            self.comment = id3["comment"].value
            self.author = id3["author"].value
            self.title = id3["song"].value
            self.album = id3["album"].value
            self.creation_date = id3["year"].value
        if "id3v2" in mp3:
            self.readID3v2(mp3["id3v2"])
        if "music" in mp3:
            # Guess music duration using fixed bit rate
            self.duration = (mp3["music"].size * 1000) / self.bits_per_sec

registerExtractor(AuFile, AuMetadata)
registerExtractor(Mp3File, Mp3Metadata)

