from hachoir.metadata.metadata import (Metadata, MultipleMetadata,
    registerExtractor)
from hachoir.parser.image import (
    BmpFile, IcoFile, PcxFile, GifFile, PngFile, XcfFile)
from hachoir.parser.image.xcf import XcfProperty

class ImageMetadata(Metadata):
    header = "Image"
    def __init__(self):
        Metadata.__init__(self)
        self.register("width", 311, "Image width")
        self.register("height", 312, "Image height")
        self.register("image_orientation", 313, "Image orientation")

        self.register("nb_colors", 315, "Number of colors")
        self.register("bits_per_pixel", 316, "Bits/pixel")
        self.register("image_format", 317, "Image format")

class BmpMetadata(ImageMetadata):
    def extract(self, bmp):
        self.width = bmp["width"].value
        self.height = bmp["height"].value
        bpp = bmp["bpp"].value
        if bpp <= 8 and "used_colors" in bmp:
            self.nb_colors = bmp["used_colors"].value
        self.bits_pixel_pixel = bpp
        self.compression = bmp["compression"].display

class IcoMetadata(MultipleMetadata):
    color_to_bpp = {
        2: 1,
        16: 4,
        256: 8
    }

    def extract(self, icon):
        index = 0
        while ("icon_data[%u]" % index) in icon:
            image = ImageMetadata()
            header = icon["icon_header[%u]" % index]
            header2 = icon["icon_data[%u]/header" % index]

            image.width = header["width"].value
            image.height = header["height"].value
            bpp = header["bpp"].value
            image.compression = header2["compression"].display
            nb_colors = header["nb_color"].value
            if nb_colors != 0:
                image.nb_colors = nb_colors
                if bpp == 0 and nb_colors in self.color_to_bpp:
                    bpp = image.color_to_bpp[nb_colors]
            image.bits_per_pixel = bpp
            image.setHeader("Icon #%u (%sx%s)"
                % (1+index, image.width, image.height))
            self.addGroup("image[%u]" % index, image)
            index += 1

class PcxMetadata(ImageMetadata):
    def extract(self, pcx):
        self.width = 1 + pcx["xmax"].value
        self.height = 1 + pcx["ymax"].value
        self.bits_per_pixel = pcx["bpp"].value
        self.compression = "Run-length encoding (RLE)"
        self.format_version = "PCX: %s" % pcx["version"].display

class XcfMetadata(ImageMetadata):
    # Map image type to bits/pixel
    type_to_bpp = {
        0: 24,
        1: 8,
        2: 8
    }

    def processProperty(self, prop):
        type = prop["type"].value
        if type == XcfProperty.PROP_PARASITES:
            for field in prop["data"]:
                if field["name"].value == "gimp-comment":
                    self.comment = field["data"].value
        elif type == XcfProperty.PROP_COMPRESSION:
            self.compression = prop["data/compression"].display

    def readProperties(self, xcf):
        found = False
        for field in xcf:
            if field.name.startswith("property["):
                found = True
                self.processProperty(field)
            elif found:
                return

    def extract(self, xcf):
        self.width = xcf["width"].value
        self.height = xcf["height"].value
        self.bits_per_pixel = self.type_to_bpp[ xcf["type"].value ]
        self.format_version = xcf["type"].display
        self.readProperties(xcf)

class PngMetadata(ImageMetadata):
    def extract(self, png):
        header = png["/header"]
        self.width = header["width"].value
        self.height = header["height"].value
        bpp = header["bpp"].value
        if header["palette"].value:
            self.nb_colors = png["/palette/size"].value / 3
        if 24 <= bpp:
            if header["alpha"].value:
                self.image_format = "RGBA"
            else:
                self.image_format = "RGB"
        else:
            self.image_format = "Color index"
        self.bits_per_pixel = bpp
        self.compression = header["compression"].display
        if "time" in png:
            self.creation_date = str(png["time"].value)
        for field in png:
            if field.name.startswith("text"):
                keyword = field["keyword"].value
                text = field["text"].value
                if keyword.lower() != "comment":
                    self.comment = "%s=\"%s\"" % (keyword, text)
                else:
                    self.comment = text

class GifMetadata(ImageMetadata):
    def extract(self, gif):
        header = gif["/screen"]
        self.width = header["width"].value
        self.height = header["height"].value
        self.bits_per_pixel = (1 + header["bpp"].value)
        self.nb_colors = (1 << self.bits_per_pixel)
        self.compression = "LZW"
        self.format_version =  "GIF version %s" % gif["header"].value[-3:]

registerExtractor(IcoFile, IcoMetadata)
registerExtractor(GifFile, GifMetadata)
registerExtractor(XcfFile, XcfMetadata)
registerExtractor(PcxFile, PcxMetadata)
registerExtractor(BmpFile, BmpMetadata)
registerExtractor(PngFile, PngMetadata)

