from hachoir.error import HachoirError
from hachoir.metadata.metadata import Metadata, MultipleMetadata, registerExtractor
from hachoir.metadata.image import ImageMetadata
from hachoir.parser.image.exif import ExifEntry
from hachoir.parser.image.jpeg import JpegFile
from hachoir import config
import types

class JpegMetadata(ImageMetadata):
    exif_key = {
        # Exif metadatas
        ExifEntry.TAG_CAMERA_MANUFACTURER: "camera_manufacturer",
        ExifEntry.TAG_CAMERA_MODEL: "camera_model",
        ExifEntry.TAG_ORIENTATION: "image_orientation",
        ExifEntry.TAG_EXPOSURE: "camera_exposure",
        ExifEntry.TAG_FOCAL: "camera_focal",
        ExifEntry.TAG_BRIGHTNESS: "camera_brightness",
        ExifEntry.TAG_APERTURE: "camera_aperture",

        # Generic metadatas
        ExifEntry.TAG_IMG_TITLE: "title",
        ExifEntry.TAG_SOFTWARE: "producer",
        ExifEntry.TAG_FILE_TIMESTAMP: "creation_date",
        ExifEntry.TAG_WIDTH: "width",
        ExifEntry.TAG_HEIGHT: "height"
    }

    iptc_key = {
        120: "title",
        116: "copyright",
#            231: "comments"
    }


    orientation_name = {
        1: 'Horizontal (normal)',
        2: 'Mirrored horizontal',
        3: 'Rotated 180',
        4: 'Mirrored vertical',
        5: 'Mirrored horizontal then rotated 90 CCW',
        6: 'Rotated 90 CW',
        7: 'Mirrored horizontal then rotated 90 CW',
        8: 'Rotated 90 CCW'
    }

    def __init__(self):
        ImageMetadata.__init__(self)
        self.register("camera_aperture", 500, "Camera aperture")
        self.register("camera_focal", 501, "Camera focal")
        self.register("camera_exposure", 502, "Camera exposure")
        self.register("camera_brightness", 510, "Camera brightness")
        self.register("camera_model", 520, "Camera model")
        self.register("camera_manufacturer", 521, "Camera manufacturer")

    def extract(self, jpeg):
        if "sof/content" in jpeg:
            sof = jpeg["sof/content"]
            self.width = sof["width"].value
            self.height = sof["height"].value
            nb_components = sof["nr_components"].value
            self.bits_per_pixel = 8 * nb_components
            if nb_components == 3:
                self.image_format = "YCbCr"
            else:
                self.image_format = "Grayscale"
                self.nb_colors = 256
        self.compression = "JPEG"
        if "app0/content" in jpeg:
            app0 = jpeg["app0/content"]
            self.format_version = "JPEG %s.%s" \
                % (app0["ver_maj"].value, app0["ver_min"].value)

        if "exif/content" in jpeg:
            for field in jpeg["exif/content"]:
                if field.name.startswith("ifd"):
                    self.parseExifIFD(field)
        if "psd/content" in jpeg:
            psd = jpeg["psd/content"]
            if "iptc" in psd:
                self.parseIPTC(psd["iptc/content"], info)

    def parseExifIFD(self, ifd):
        for field in ifd:
            # Don't process EXIF values
            if field.name == "next":
                break

            # Only process entries
            if not field.name.startswith("entry"):
                continue

            # Skip unknown tags
            tag = field["tag"].value
            if tag not in self.exif_key:
                continue
            key = self.exif_key[tag]

            # Read value
            rational = False
            if "value" in field:
                value = field["value"].value
            elif field["type"].value in (ExifEntry.TYPE_RATIONAL, ExifEntry.TYPE_SIGNED_RATIONAL):
                value  = float(ifd[ "value_%s[0]" % field.name ].value)
                value /= ifd[ "value_%s[1]" % field.name ].value
                rational = True
            else:
                value = ifd["value_%s" % field.name].value

            # Convert value to string
            if tag == ExifEntry.TAG_ORIENTATION:
                value = self.orientation_name.get(value, value)
            elif tag == ExifEntry.TAG_EXPOSURE:
                value = "1/%g" % (1/value)
            elif rational:
                value = "%.3g" % value

            # Store information
            setattr(self, key, value)

    def parseIPTC(self, iptc, info):
        for field in iptc:
            # Skip incomplete field
            if "tag" not in field or "content" not in field:
                continue

            # Skip unknown tag
            tag = field["tag"].value
            if tag not in self.iptc_key:
                continue
            key = self.iptc_key[tag]

            # Get value
            value = field["content"].value
            if isinstance(value, types.StringTypes):
                value = value.replace("\r", " ")
                value = value.replace("\n", " ")
            setattr(self, key, value)
        if config.debug:
            self.dumpIPTC(iptc)

    def dumpIPTC(self, iptc):
        print "--- IPTC tags ---"
        for field in iptc:
            desc = field.description
            if not desc:
                desc = "-"
            if "content" in field:
                content = field["content"].display
            else:
                content = "-"
            print "% 18s | % 3s | % 35s | %s" \
                % (field.name, field["tag"].value, desc, content)

registerExtractor(JpegFile, JpegMetadata)

