from hachoir.error import error, HachoirError
from hachoir.endian import endian_name
from hachoir.tools import humanDuration, makePrintable
from hachoir.dict import Dict
import types

extractors = {}

class Data:
    def __init__(self, key, priority, description,
    repetable=False, handler=None):
        assert Metadata.MIN_PRIORITY <= priority <= Metadata.MAX_PRIORITY
        self.key = key
        self.description = description
        self.repetable = repetable
        self.values = []
        self.handler = handler
        self.priority = priority

    def __contains__(self, value):
        return value in self.values

    def __cmp__(self, other):
        return cmp(self.priority, other.priority)

class Metadata(object):
    MIN_PRIORITY = 100
    MAX_PRIORITY = 999
    header = "Metadata"

    def __init__(self):
        object.__init__(self)
        object.__setattr__(self, "_Metadata__data", {})
        header = self.__class__.header
        object.__setattr__(self, "_Metadata__header", header)
        self.register("title", 100, "Title")
        self.register("author", 101, "Author", True)
        self.register("duration", 200, "Duration", handler=humanDuration) # type: int, in millisec

        self.register("creation_date", 500, "Creation date", True)
        self.register("last_modification", 501, "Last modification", True)

        self.register("compression", 600, "Compression")
        self.register("copyright", 601, "Copyright", True)

        self.register("url", 900, "URL", True)
        self.register("comment", 901, "Comment", True)
        self.register("producer", 902, "Producer", True)
        self.register("format_version", 950, "Format version")
        self.register("mime_type", 951, "MIME type")
        self.register("endian", 952, "Endian")

    def __setattr__(self, key, value):
        """
        Add a new value to data with name 'key'. Skip value if it does already
        exist. Generate an HachoirError when trying to store two value in a
        non-repetable data.
        """
        if key not in self.__data:
            raise KeyError("%s has no metadata '%s'" % (self.__class__.__name__, key))

        # Don't store empty string
        if isinstance(value, types.StringTypes) and not value:
            return

        data = self.__data[key]
        if not data.repetable:
            if value in data:
                raise HachoirError("Attribute %s (%s) is not repetable!" \
                    % (key, data.description))
            data.values.append(value)
        elif value not in data:
            data.values.append(value)

    def setHeader(self, text):
        object.__setattr__(self, "_Metadata__header", text)

    def __getattr__(self, key):
        """
        Read values of tag with name 'key'. Values may be an empty list.
        """
        if key not in self.__data:
            raise AttributeError("%s has no attribute '%s'"
                % (self.__class__.__name__, key))
        data = self.__data[key]
        if data.repetable:
            return self.__data[key].values
        elif len(data.values) == 1:
            return self.__data[key].values[0]
        else:
            raise AttributeError("Attribute '%s' of %s is not set"
                % (key, self.__class__.__name__))

    def register(self, key, priority, title, repetable=False, handler=None):
        assert key not in self.__data
        self.__data[key] = Data(key, priority, title, repetable, handler)

    def __iter__(self):
        return self.__data.itervalues()

    def __str__(self):
        """
        Create a multi-line ASCII string which represent all datas.

        >>> a = Metadata()
        >>> a.author = "haypo"
        >>> a.author = "jm"
        >>> print a
        Metadata:
        - Author: haypo
        - Author: jm

        @see exportPlaintext()
        """
        return self.exportPlaintext(self.MAX_PRIORITY, "ascii")

    def exportPlaintext(self, priority, charset, human=True):
        """
        Create a multi-line string in specified charset which represent all
        datas. priority is the maximum accepted priority.

        @see __str__()
        """
        text = []
        for data in sorted(self):
            if priority < data.priority:
                break
            for value in data.values:
                if data.handler:
                    value = data.handler(value)
                value = unicode(value)
                value = makePrintable(value, charset)
                if human:
                    prefix = data.description
                else:
                    prefix = data.key
                text.append("%s: %s" % (prefix, value))
        if not text:
            return None
        text = "\n- ".join(text)
        return "%s:\n- %s" % (self.__header, text)

class MultipleMetadata(Metadata):
    header = "Common"
    def __init__(self):
        Metadata.__init__(self)
        object.__setattr__(self, "_MultipleMetadata__groups", Dict())

    def __contains__(self, key):
        return key in self.__groups

    def __getitem__(self, key):
        return self.__groups[key]

    def addGroup(self, key, metadata, header=None):
        if header:
            metadata.setHeader(header)
        self.__groups.append(key, metadata)

    def exportPlaintext(self, priority, charset, human=True):
        common = Metadata.exportPlaintext(self, priority, charset, human)
        if common:
            text = [ common ]
        else:
            text = []
        for metadata in self.__groups:
            value = metadata.exportPlaintext(priority, charset, human)
            if value:
                text.append(value)
        if not text:
            return None
        return "\n".join(text)

def registerExtractor(parser, extractor):
    assert parser not in extractors
    extractors[parser] = extractor

def extractMetadata(field_set):
    """
    Create a Metadata class from a FieldSet instance.
    Returns None if we don't know about this format
    """
    parser = field_set.__class__
    if parser not in extractors:
        return None
    extractor = extractors[parser]
    metadata = extractor()
    metadata.extract(field_set)
    metadata.mime_type = field_set.mime_type
    metadata.endian = endian_name[field_set.endian]
    return metadata

