"""
Extract metadata from RIFF file format: AVI video and WAV sound.
"""

from hachoir.metadata.metadata import Metadata, MultipleMetadata, registerExtractor
from hachoir.metadata.audio import AudioMetadata
from hachoir.metadata.video import VideoMetadata
from hachoir.parser.video.riff import RiffFile
from hachoir.parser.video.fourcc import AUDIO_MICROSOFT_PCM
from hachoir.tools import humanDuration, humanFilesize

class WavMetadata(AudioMetadata):
    def extract(self, wav):
        format = wav["format"]
        self.sample_rate = format["sample_per_sec"].value
        self.nb_channel = format["nb_channel"].value
        self.compression = format["codec"].display
        if "fact" in wav:
            self.duration = wav["fact/content/nb_samples"].value / (self.sample_rate / 1000)
        elif format["codec"].value == AUDIO_MICROSOFT_PCM:
            nb_frame = (self.sample_rate * format["bit_per_sample"].value / 8) / 1000
            self.duration = wav["audio_data/size"].value / nb_frame
        if "audio_data" in wav and hasattr(self, "duration"):
            self.bits_per_sec = (wav["audio_data"].size * 1000) / self.duration

class RiffMetadata(MultipleMetadata):
    tag_to_key = {
        "INAM": "title",
        "IART": "artist",
        "ICMT": "comment",
        "ICOP": "copyright",
        "IENG": "author",
        "ISFT": "producer",
        "ICRD": "creation_date"
    }

    def processChunk(self, chunk):
        tag = chunk["tag"].value
        if tag not in self.tag_to_key:
            return
        key = self.tag_to_key[tag]
        value = chunk["text"].value
        setattr(self, key, value)

    def extract(self, riff):
        type = riff["type"].value
        if type == "WAVE":
            wav = WavMetadata()
            wav.extract(riff)
            self.addGroup("audio", wav)
        elif type == "AVI ":
            self.extractAVI(riff)

        self.extractInfo(riff)

    def extractInfo(self, fieldset):
        for field in fieldset:
            if not field.is_field_set:
                continue
            if "tag" in field:
                if field["tag"].value == "LIST" \
                and field["subtag"].value != "movi":
                    self.extractInfo(field)
                else:
                    self.processChunk(field)

    def extractAVI(self, avi):
        metav = VideoMetadata()
        metaa = AudioMetadata()

        # Get headers
        headers = avi["headers"]
        video = headers["stream[0]"]
        videoh = video["stream_hdr"]
        videof = video["stream_fmt"]
        audio = headers["stream[1]"]
        audioh = audio["stream_hdr"]
        audiof = audio["stream_fmt"]

        # Extract video information
        metav.compression = "%s (fourcc:\"%s\")" % (videoh["fourcc"].display, videoh["fourcc"].value)
        metav.frame_rate = float(videoh["rate"].value) / videoh["scale"].value

        # Video size
        if "v_width" in videof:
            metav.width = videof["v_width"].value
            metav.height = videof["v_height"].value
            metav.bits_per_pixel = videof["v_depth"].value
        else:
            metav.width = videoh["right"].value - videoh["left"].value
            metav.height = videoh["bottom"].value - videoh["top"].value

        # Video has index?
        if "index" in avi:
            metav.comment = "Has index (%s)" \
                % humanFilesize(avi["index"].size/8)

        # Extract audio informations
        metaa.nb_channel = audiof["channel"].value
        metaa.sample_rate = audiof["sample_rate"].value
        if audioh["fourcc"].value != "":
            metaa.compression = "%s (fourcc:\"%s\")" % (audiof["codec"].display, audioh["fourcc"].display)
        else:
            metaa.compression = audiof["codec"].display

        # Common attributes
        self.duration = videoh["length"].value * 1000 / metav.frame_rate

        # Save data
        self.addGroup("video", metav)
        self.addGroup("audio", metaa)


registerExtractor(RiffFile, RiffMetadata)

