from hachoir.error import HachoirError
from hachoir.metadata.metadata import MultipleMetadata, registerExtractor
from hachoir.metadata.image import ImageMetadata
from hachoir.metadata.audio import AudioMetadata
from hachoir.parser.video import (MovFile, MkvFile)

class VideoMetadata(ImageMetadata):
    header = "Video"

    def __init__(self):
        ImageMetadata.__init__(self)
        self.register("artist", 321, "Artist", True)
        self.register("subtitle_author", 421, "Subtitle author", True)
        self.register("frame_rate", 620, "Frame rate")

class MkvMetadata(MultipleMetadata):
    tag_key = {
        "TITLE": "title",
        "URL": "url",
#        "SUBTITLE": "subtitle_author",
        "COPYRIGHT": "copyright",
        "DATE_RECORDED": "creation_date"  # use maybe another name?
    }

    def extract(self, mkv):
        if "Segment[0]" not in mkv:
            raise HachoirError("Invalid Matoska video")
        segment_index = 0
        while True:
            self.processSegment( mkv["Segment[%s]" % segment_index] )
            segment_index += 1
            if ("Segment[%s]" % segment_index) not in mkv:
                break

    def processSegment(self, segment):
        for field in segment:
            if field.name.startswith("Info["):
                self.processInfo(field)
            elif field.name.startswith("Tags["):
                self.processTags(field)
            elif field.name.startswith("Tracks["):
                self.processTracks(field)

    def processTracks(self, tracks):
        for field in tracks:
            if field.name.startswith("TrackEntry["):
                self.processTrack(field)

    def processTrack(self, track):
        if track["TrackType/enum"].display == "video":
            self.processVideo(track)
        elif track["TrackType/enum"].display == "audio":
            self.processAudio(track)

    def processVideo(self, track):
        video = VideoMetadata()
        video.width = track["Video/PixelWidth/unsigned"].value
        video.height = track["Video/PixelHeight/unsigned"].value
        video.compression = track["CodecID/string"].value
        self.addGroup("video", video)

    def processAudio(self, track):
        audio = AudioMetadata()
        audio.sample_rate = track["Audio/SamplingFrequency/float"].value
        audio.nb_channel = track["Audio/Channels/unsigned"].value
        self.addGroup("audio", audio)

    def processTags(self, tags):
        for field in tags:
            if field.name.startswith("Tag["):
                self.processTag(field)

    def processTag(self, tag):
        for field in tag:
            if field.name.startswith("SimpleTag["):
                self.processSimpleTag(field)

    def processSimpleTag(self, tag):
        name = tag["TagName/unicode"].value
        if name not in self.tag_key:
            return
        key = self.tag_key[name]
        value = tag["TagString/unicode"].value
        setattr(self, key, value)

    def processInfo(self, info):
        if "Duration" in info and "TimecodeScale" in info:
            self.duration = (info["Duration/float"].value * info["TimecodeScale/unsigned"].value) / 1000000
        if "DateUTC" in info:
            self.creation_date = info["DateUTC/date"].display
        producer = []
        if "WritingApp" in info:
            producer.append(info["WritingApp/unicode"].value)
        if "MuxingApp" in info:
            producer.append(info["MuxingApp/unicode"].value)
        if producer:
            self.producer = '; '.join(producer)

class MovMetadata(VideoMetadata):
    def extract(self, mov):
        for atom in mov:
            if "movie" in atom:
                self.processMovie(atom["movie"])

    def processMovieHeader(self, hdr):
        self.frame_rate = hdr["time_scale"].value
        self.duration = hdr["duration"].value * 1000 / self.frame_rate
        self.creation_date = hdr["creat_date"].display
        self.last_modification = hdr["lastmod_date"].display
        self.comment = "Play speed: %.1f%%" % (hdr["play_speed"].value*100)
        self.comment = "User volume: %.1f%%" % (float(hdr["volume"].value)*100//255)

    def processMovie(self, atom):
        for field in atom:
            if "movie_hdr" in field:
                self.processMovieHeader(field["movie_hdr"])

registerExtractor(MovFile, MovMetadata)
registerExtractor(MkvFile, MkvMetadata)

