"""
GZIP archive parser.

Author: Victor Stinner
"""

from hachoir.field import (Parser,
    UInt8, UInt16, UInt32, Enum,
    Bit, CString,
    RawBits, Bytes, RawBytes)
from hachoir.text_handler import hexadecimal, humanFilesize, timestampUNIX
from hachoir.endian import LITTLE_ENDIAN

class GzipFile(Parser):
    endian = LITTLE_ENDIAN
    tags = {
        "file_ext": "gz",
        "mime": ["application/x-gzip"],
        "min_size": 18*8,
        "description": "gzip archive"
    }
    os_name = {
        0: "FAT filesystem",
        1: "Amiga",
        2: "VMS (or OpenVMS)",
        3: "Unix",
        4: "VM/CMS",
        5: "Atari TOS",
        6: "HPFS filesystem (OS/2, NT)",
        7: "Macintosh",
        8: "Z-System",
        9: "CP/M",
        10: "TOPS-20",
        11: "NTFS filesystem (NT)",
        12: "QDOS",
        13: "Acorn RISCOS"}

    def validate(self):
        return (self["signature"].value == '\x1F\x8B')

    def createFields(self):
        # Gzip header
        yield Bytes(self, "signature", 2, r"GZip file signature (\x1F\x8B)")
        yield UInt8(self, "compression", "Compression method",
            text_handler=self.getCompressionMethod)

        # Flags
        yield Bit(self, "is_text", "File content is probably ASCII text")
        yield Bit(self, "has_crc16", "Header CRC16")
        yield Bit(self, "has_extra", "Extra informations (variable size)")
        yield Bit(self, "has_filename", "Contains filename?")
        yield Bit(self, "has_comment", "Contains comment?")
        yield RawBits(self, "reserved[]", 3)
        yield UInt32(self, "mtime", "Modification time",
            text_handler=timestampUNIX)

        # Extra flags
        yield RawBits(self, "reserved[]", 1)
        yield Bit(self, "slowest", "Compressor used maximum compression (slowest)")
        yield Bit(self, "fastest", "Compressor used the fastest compression")
        yield RawBits(self, "reserved[]", 5)
        yield Enum(UInt8(self, "os", "Operating system"), GzipFile.os_name)

        # Optional fields
        if self["has_extra"].value:
            yield UInt16(self, "extra_length", "Extra length")
            yield RawBytes(self, "extra", self["extra_length"].value, "Extra")
        if self["has_filename"].value:
            yield CString(self, "filename", "Filename")
        if self["has_comment"].value:
            yield CString(self, "comment", "Comment")
        if self["has_crc16"].value:
            yield UInt16(self, "hdr_crc16", "CRC16 of the header",
                text_handler=hexadecimal)

        # Read content
        size = (self.size - self.current_size) / 8 - 8  # -8: crc32+size
        yield RawBytes(self, "data", size, "Compressed data")

        # Footer
        yield UInt32(self, "crc32", "Uncompressed data content CRC32",
            text_handler=hexadecimal)
        yield UInt32(self, "size", "Uncompressed size",
            text_handler=humanFilesize)

    def getCompressionMethod(self, chunk):
        if chunk.value < 8:
            return "reserved"
        elif chunk.value == 8:
            return "deflate"
        else:
            return "Unknow (%s)" % chunk.value

