"""
Tar archive parser.

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, Parser,
    Enum, UInt8, RawBytes, String, PaddingBytes)
from hachoir.error import error
from hachoir.tools import humanFilesize, paddingSize, humanUnixAttributes
from hachoir.endian import BIG_ENDIAN
import datetime
import re

class FileEntry(FieldSet):
    type_name = {
        0: "Normal disk file (old format)",
        # 48 is "0", 49 is "1", ...
        48: "Normal disk file",
        49: "Link to previously dumped file",
        50: "Symbolic link",
        51: "Character special file",
        52: "Block special file",
        53: "Directory",
        54: "FIFO special file",
        55: "Contiguous file"
    }

    def getOctal(self, name):
        return self.octal2int(self[name].value)

    def getDatetime(self):
        timestamp = self.getOctal("mtime")
        return datetime.datetime.fromtimestamp(timestamp)

    def createFields(self):
        yield String(self, "name", 100, "Name", strip="\0") # TODO: charset?
        yield String(self, "mode", 8, "Mode", strip=" \0", charset="ascii")
        yield String(self, "uid", 8, "User ID", strip=" \0", charset="ascii")
        yield String(self, "gid", 8, "Group ID", strip=" \0", charset="ascii")
        yield String(self, "size", 12, "Size", strip=" \0", charset="ascii")
        yield String(self, "mtime", 12, "Modification time", strip=" \0", charset="ascii")
        yield String(self, "check_sum", 8, "Check sum", strip=" \0", charset="ascii")
        yield Enum(UInt8(self, "type", "Type"), self.type_name)
        yield String(self, "lname", 100, "Link name", strip=" \0") # TODO: charset?
        yield String(self, "magic", 8, "Magic", strip=" \0", charset="ascii")
        yield String(self, "uname", 32, "User name", strip=" \0") # TODO: charset?
        yield String(self, "gname", 32, "Group name", strip=" \0") # TODO: charset?
        yield String(self, "devmajor", 8, "Dev major", strip=" \0", charset="ascii")
        yield String(self, "devminor", 8, "Dev minor", strip=" \0", charset="ascii")
        yield PaddingBytes(self, "padding", 167, "Padding (zero)", pattern="\0")

        filesize = self.getOctal("size")
        if filesize != 0 and self["type"].value in (0, ord("0")):
            yield RawBytes(self, "content", filesize, "File content")

        size = paddingSize(self.current_size/8, 512)
        if size != 0:
            yield PaddingBytes(self, "padding_end", size, "Padding (512 align)", pattern="\0")

#    def convertMode(self, chunk):
#        mode = self.octal2int(chunk.value)
#        return humanUnixAttributes(mode)

    def convertOctal(self, chunk):
        return self.octal2int(chunk.value)

    def getTime(self, chunk):
        value = self.octal2int(chunk.value)
        return str(datetime.datetime.fromtimestamp(value))

    def isEmpty(self):
        return self["name"].value == ""

    def octal2int(self, text):
        try:
            return int(text, 8)
        except ValueError:
            return 0

    def _createDescription(self):
        if self.isEmpty():
            desc = "(terminator, empty header)"
        else:
            filename = self["name"].value
            filesize = humanFilesize(self.getOctal("size"))
            desc = "(%s: %s, %s)" % \
                (filename, self["type"].display, filesize)
        return "Tar File " + desc

    def updateParent(self, chunk):
        chunk.description = text
        self.setDescription(text)

class TarFile(Parser):
    endian = BIG_ENDIAN
    tags = {
        "file_ext": "tar",
        "mime": ("application/x-tar", "application/x-gtar"),
        "min_size": 512*8,
        "description": "TAR archive"
    }

    def validate(self):
        if self.stream.readBytes(257*8, 5) != "ustar":
            return "Invalid magic number"
        try:
            int(self["file[0]/uid"].value, 8)
            int(self["file[0]/gid"].value, 8)
            #int(self["file[0]/mode"].value, 8)
            int(self["file[0]/size"].value, 8)
        except ValueError:
            return "Invalid file size"
        return True

    def createFields(self):
        while self.current_size < self.size:
            field = FileEntry(self, "file[]")
            yield field
            if field.isEmpty():
                break
        padding = (self.size - self.current_size) / 8
        if padding:
            yield PaddingBytes(self, "end", padding, "End of file")

