"""
AU audio file parser

Author: Victor Stinner
Creation: 12 july 2006
"""

from hachoir.field import (FieldSet, Parser,
    UInt32, Enum, String, RawBytes)
from hachoir.endian import BIG_ENDIAN
from hachoir.text_handler import humanFilesize, humanFrequency

class AuFile(Parser):
    tags = {
        "file_ext": ["au", "snd"],
        "mime": ["audio/basic"],
        "min_size": 24*8,
        "description": "Sun/NeXT audio"
    }
    endian = BIG_ENDIAN

    codec_info = {
        1: (8, "8-bit ISDN u-law"),
        2: (8, "8-bit linear PCM [REF-PCM]"),
        3: (16, "16-bit linear PCM"),
        4: (24, "24-bit linear PCM"),
        5: (32, "32-bit linear PCM"),
        6: (32, "32-bit IEEE floating point"),
        7: (64, "64-bit IEEE floating point"),
        23: (None, "8-bit ISDN u-law compressed (CCITT G.721 ADPCM)")
    }

    # Create bit rate and codec name dictionnaries
    bit_rate = dict(( (key, value[0]) for key,value in codec_info.iteritems() ))
    codec_name = dict(( (key, value[1]) for key,value in codec_info.iteritems() ))

    def validate(self):
        if self.stream.readBytes(0, 4) != ".snd":
            return "Wrong file signature"
        return True

    def getBitRate(self):
        """
        Get bit rate (number of bit per sample per channel),
        may returns None if you unable to compute it.
        """
        return self.bit_rate.get(self["codec"].value)

    def createFields(self):
        yield String(self, "signature", 4, 'Format signature (".snd")', charset="ascii")
        yield UInt32(self, "data_ofs", "Data offset")
        yield UInt32(self, "data_size", "Data size", text_handler=humanFilesize)
        yield Enum(UInt32(self, "codec", "Audio codec"), self.codec_name)
        yield UInt32(self, "sample_rate", "Number of samples/second", text_handler=humanFrequency)
        yield UInt32(self, "channels", "Number of interleaved channels")

        size = self["data_ofs"].value - self.current_size/8
        if 0 < size:
            yield String(self, "info", size, "Information", strip=" \0") # TODO: charset?

        size = (self._size - self.current_size)/8
        yield RawBytes(self, "audio_data", size, "Audio data")

    def _createDescription(self):
        return "Sun/NeXT audio: %s" % self.getBitRate()

