"""
iPod iTunesDB parser.

Documentation:
- http://ipodlinux.org/ITunesDB

Author: Romain HERAULT
Creation date: 19 august 2006
"""

from hachoir.field import (FieldSet, Parser, ParserError,
    UInt8, UInt16, UInt32, UInt64,
    String, Float32, RawBytes, PaddingBytes)
from hachoir.endian import LITTLE_ENDIAN
from hachoir.text_handler import (timestampMac,
    humanFilesize, humanDuration)

class DataObjectHeader(FieldSet):
    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value *8

    def createFields(self):
        yield String(self, "header_id", 4, "Data Object Header Markup (\"mhod\")", charset="latin-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "type", "type")
        if(self["type"].value<15):
            yield UInt32(self, "unknow[]")
            yield UInt32(self, "unknow[]")
            yield UInt32(self, "position", "Position")
            yield UInt32(self, "length", "String Length in bytes")
            yield UInt32(self, "unknow[]")
            yield UInt32(self, "unknow[]")
            yield String(self, "string", self["length"].value/2, "String Data", charset="UTF-16LE")
            #yield RawBytes(self, "string", self["length"].value, "String Data")
        else:
            size = self["header_length"].value-12
            if size>0:
                yield RawBytes(self, "header padding", size)
        size = self["entry_length"].value - self.current_size/ 8
        if size>0:
            yield RawBytes(self, "entry padding", size)


class TrackItemHeader(FieldSet):
    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value *8

    def createFields(self):
        yield String(self, "header_id", 4, "Track Item Header Markup (\"mhit\")", charset="latin-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "string_number", "Number of Strings")
        yield UInt32(self, "unique_id", "Unique ID")
        yield UInt32(self, "visible_tag", "Visible Tag")
        yield String(self, "file_type", 4, "File Type")
        yield UInt8(self, "x1_type", "Extended Type 1")
        yield UInt8(self, "x2_type", "Extended type 2")
        yield UInt8(self, "compilation_flag", "Compilation Flag")
        yield UInt8(self, "rating", "Rating")
        yield UInt32(self, "added_date", "Date when the item was added", text_handler=timestampMac)
        yield UInt32(self, "size", "Track size in bytes", text_handler=humanFilesize)
        yield UInt32(self, "length", "Track length in milliseconds", text_handler=humanDuration)
        yield UInt32(self, "track_number", "Number of this track")
        yield UInt32(self, "total_track", "Total number of tracks")
        yield UInt32(self, "year", "Year of the track")
        yield UInt32(self, "bitrate", "Bitrate")
        yield UInt32(self, "samplerate", "Sample Rate")
        yield UInt32(self, "volume", "volume")
        yield UInt32(self, "start_time", "Start playing at, in milliseconds")
        yield UInt32(self, "stop_time", "Stop playing at,  in milliseconds")
        yield UInt32(self, "soundcheck", "SoundCheck preamp")
        yield UInt32(self, "playcount_1", "Play count of the track")
        yield UInt32(self, "playcount_2", "Play count of the track (identical to playcount_1)")
        yield UInt32(self, "last_played_time", "Time the song was last played")
        yield UInt32(self, "disc_number", "disc number in multi disc sets")
        yield UInt32(self, "total_discs", "Total number of discs in the disc set")
        yield UInt32(self, "userid", "User ID in the DRM scheme")
        yield UInt32(self, "last_modified", "Time of the last modification of the track", text_handler=timestampMac)
        yield UInt32(self, "bookmark_time", "Bookmark time for AudioBook")
        yield UInt64(self, "dbid", "Unique DataBase ID for the song (identical in mhit and in mhii)")
        yield UInt8(self, "checked", "song is checked")
        yield UInt8(self, "application_rating", "Last Rating before change")
        yield UInt16(self, "BPM", "BPM of the track")
        yield UInt16(self, "artwork_count", "number of artworks fo this item")
        yield UInt16(self, "unknow[]")
        yield UInt32(self, "artwork_size", "Total size of artworks in bytes")
        yield UInt32(self, "unknow[]")
        yield Float32(self, "sample_rate_2", "Sample Rate express in float")
        yield UInt32(self, "released_date", "Date of release in Music Store or in Podcast")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt8(self, "has_artwork", "0x01 for track with artwork, 0x02 otherwise")
        yield UInt8(self, "skip_wen_shuffling", "Skip that track when shuffling")
        yield UInt8(self, "remember_playback_position", "Remember playback position")
        yield UInt8(self, "flag4", "Flag 4")
        yield UInt64(self, "dbid2", "Unique DataBase ID for the song (identical as above)")
        yield UInt8(self, "lyrics_flag", "Lyrics Flag")
        yield UInt8(self, "movie_file_flag", "Movie File Flag")
        yield UInt8(self, "played_mark", "Track has been played")
        yield UInt8(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "sample_count", "Number of samples in the song (only for WAV and AAC files)")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "media_type", "Media Type for video iPod")
        yield UInt32(self, "season_number", "Season Number")
        yield UInt32(self, "episode_number", "Episode Number")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        yield UInt32(self, "unknow[]")
        size = self["header_length"].value-244
        if size>0:
            yield RawBytes(self, "header padding", size)

        #while ((self.stream.readBytes(0, 4) == 'mhod') and  ((self.current_size/8) < self["entry_length"].value)):
        for i in xrange(self["string_number"].value):
            yield DataObjectHeader(self, "data[]")
        size = self["entry_length"].value - self.current_size/ 8
        if size>0:
            yield RawBytes(self, "entry padding", size)

class TrackListHeader(FieldSet):
    def createFields(self):
        yield String(self, "header_id", 4, "Track List Header Markup (\"mhlt\")", charset="latin-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "track_number", "Number of Tracks")

        size = self["header_length"].value-12
        if size>0:
            yield RawBytes(self, "header padding", size)

        for i in xrange(self["track_number"].value):
                yield TrackItemHeader(self, "track[]")

class DataSetHeader(FieldSet):
    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value *8

    def createFields(self):
        yield String(self, "header_id", 4, "DataSet Header Markup (\"mhsd\")", charset="latin-1")
        yield UInt32(self, "header_length", "Header Length")
        yield UInt32(self, "entry_length", "Entry Length")
        yield UInt32(self, "type", "type")
        size = self["header_length"].value-16
        if size>0:
            yield RawBytes(self, "header_raw", size)

        if self["type"].value == 1:
            yield TrackListHeader(self, "tracklist")
        size = self["entry_length"].value - self.current_size/ 8
        if size>0:
            yield RawBytes(self, "entry padding", size)

def parseEntry(self):
    yield String(self, "header_id", 4, "DataBase Header Markup (\"mhbd\")", charset="latin-1")
    yield UInt32(self, "header_length", "Header Length")
    yield UInt32(self, "entry_length", "Entry Length")
    yield UInt32(self, "unknow[]")
    yield UInt32(self, "version_number", "Version Number")
    yield UInt32(self, "child_number", "Number of Childs")
    yield UInt64(self, "id", "ID for this database")
    yield UInt32(self, "unknow[]")
    yield UInt64(self, "initial_dbid", "Initial DBID")
    size = self["header_length"].value-(28+16)
    if size>0:
        yield PaddingBytes(self, "padding", size, pattern="\0")
    for i in xrange(self["child_number"].value):
        yield DataSetHeader(self, "dataset[]")
    size = self["entry_length"].value - self.current_size/ 8
    if size>0:
        yield RawBytes(self, "entry padding", size)

class DBHMEntry(FieldSet):
    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["entry_length"].value *8

    def createFields(self):
        for field in parseEntry(self):
            yield field

class ITunesDBFile(Parser):
    tags = {
        "min_size": 44*8,
        "description": "iPod iTunesDB file"
    }

    endian = LITTLE_ENDIAN

    def validate(self):
        return self.stream.readBytes(0, 4) == 'mhbd'

    def createFields(self):
        for field in parseEntry(self):
            yield field

