"""
MPEG audio file parser.

Creation: 12 decembre 2005
Status: alpha
Author: Victor Stinner
"""

from hachoir.field import (Parser, Parser, ParserError,
    Bit, Bits, Enum, UInt8,
    PaddingBits, PaddingBytes,
    RawBytes)
from hachoir.parser.audio.id3 import ID3v1, ID3v2
from hachoir.endian import BIG_ENDIAN

class Mp3File(Parser):
    tags = {
        "file_ext": ("mpa", "mp1", "mp2", "mp3"),
        "mime": ["audio/mpeg"],
        "min_size": 4*8,
        "description": "MPEG version 1 audio"
    }
    endian = BIG_ENDIAN

    version_name = { 0: "2.5", 2: "2", 3: "1" }
    layer_name = { 1: "III", 2: "II", 3: "I", 0: "(reserved)" }

    # Bit rates (bit_rate * 1000 = bits/sec)
    bit_rates = {
        1: ( # MPEG1
            ( 0, 32,  64,  96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448 ), # layer I
            ( 0, 32,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384 ), # layer II
            ( 0, 32,  40,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320 ), # layer III
            # -   1    2    3    4    5    6    7    8    9   10   11   12   13   14
        ),
        2: ( # MPEG2 / MPEG2.5
            ( 0, 32,  48,  56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256 ), # layer I
            ( 0,  8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160 ), # layer II
            ( 0,  8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160 ), # layer III
            # -   1    2    3    4    5    6    7    8    9   10   11   12   13   14
        )
    }
    sampling_rates = {
        3: {0: 44100, 1: 48000, 2: 32000},  # MPEG1
        2: {0: 22050, 1: 24000, 2: 16000},  # MPEG2
        0: {0: 11025, 1: 12000, 2: 8000}    # MPEG2.5
    }
    emphasis_name = {0: "none", 1: "50/15 ms",  3: "CCIT J.17"}
    channel_mode_name = {
        0: "Stereo", 1: "Joint stereo",
        2: "Dual channel", 3: "Single channel"
    }

    def validate(self):
        if self.stream.readBytes(0, 3) == "ID3":
            return True
        else:
            return (self["sync"].value == 2047)

    def createFields(self):
        if self.stream.readBytes(0, 3) == "ID3":
            yield ID3v2(self, "id3v2")

        # Header
        yield PaddingBits(self, "sync", 11, "Synchronize bits (set to 1)", pattern=1)
        yield Enum(Bits(self, "version", 2, "MPEG audio version"), self.version_name)
        yield Enum(Bits(self, "layer", 2, "MPEG audio layer"), self.layer_name)
        yield Bit(self, "crc16", "No CRC16 protection?")

        # Rates and padding
        yield Bits(self, "bit_rate", 4, "Bit rate")
        yield Bits(self, "sampling_rate", 2, "Sampling rate")
        yield Bit(self, "padding", "Stream field use padding?")
        yield Bit(self, "extension", "Extension")

        # Channel mode, mode extension, copyright, ...
        yield Enum(Bits(self, "channel_mode", 2, "Channel mode"), self.channel_mode_name)
        yield Bits(self, "mode_ext", 2, "Mode extension")
        yield Bit(self, "copyright", "Is copyrighted?")
        yield Bit(self, "original", "Is original?")
        yield Enum(Bits(self, "emphasis", 2, "Emphasis"), self.emphasis_name)

        size = (self.size - self.current_size)/8
        has_id3 = (self.stream.readBytes(self.size - 128*8, 3) == "TAG")
        if has_id3:
            size -= 128
        if size:
            yield RawBytes(self, "music", size, "Compressed music")
        if has_id3:
            yield ID3v1(self, "id3v1")

    def getSampleRate(self):
        version = self["version"].value
        rate = self["sampling_rate"].value
        return self.sampling_rates[version][rate]

    def getKBitRate(self):
        layer = 3 - self["layer"].value
        bit_rate = self["bit_rate"].value
        if self["version"].value == 3:
            return self.bit_rates[1][layer][bit_rate] # MPEG1
        else:
            return self.bit_rates[2][layer][bit_rate] # MPEG2 / MPEG2.5

    def _createDescription(self):
        sampling_rate = self.getSampleRate()
        bit_rate = self.getKBitRate()
        frame_size = (144 * sampling_rate) / bit_rate + self["padding"].value
        return "MPEG version %s, layer %s, %u kbps, %s Hz, %s" % (\
            self["version"].display, self["layer"].display,
            bit_rate, sampling_rate, self["channel_mode"].display)

