"""
Microsoft Bitmap picture parser.
- file extension: ".bmp"

Author: Victor Stinner
Creation: 16 december 2005
"""

from hachoir.field import (Parser, ParserError,
    UInt16, UInt32, String, RawBytes, Enum, PaddingBytes)
from hachoir.endian import LITTLE_ENDIAN

class BmpFile(Parser):
    tags = {
        "file_ext": "bmp",
        "mime": ("image/x-ms-bmp", "image/x-bmp"),
        "min_size": 30*8,
        "description": "Microsoft bitmap (BMP) picture"
    }
    endian = LITTLE_ENDIAN

    compression_name = {
        0: "Uncompressed",
        1: "RLE"
    }

    def validate(self):
        if self.stream.readBytes(0, 2) != 'BM':
            return "Wrong file signature"
        if self["header_size"].value not in (12, 40):
            return "Unknown header size (%s)" % self["header_size"].value
        return True

    def createFields(self):
        yield String(self, "header", 2, "Header (\"BM\")", charset="ascii")
        yield UInt32(self, "file_size", "File size (bytes)")
        yield PaddingBytes(self, "notused", 4, "Reseved", pattern="\0")
        yield UInt32(self, "data_start", "Data start position")
        yield UInt32(self, "header_size", "Header size")
        yield UInt32(self, "width", "Width (pixels)")
        yield UInt32(self, "height", "Height (pixels)")
        yield UInt16(self, "nb_plan", "Number of plan (=1)")
        yield UInt16(self, "bpp", "Bits per pixel")
        if self["header_size"].value == 40:
            yield Enum(UInt32(self, "compression", "Compression method"), BmpFile.compression_name)
            yield UInt32(self, "image_size", "Image size (bytes)")
            yield UInt32(self, "horizontal_dpi", "Horizontal DPI")
            yield UInt32(self, "vertical_dpi", "Vertical DPI")
            yield UInt32(self, "used_colors", "Number of color used")
            yield UInt32(self, "important_color", "Number of import colors")
        size = (self.size - self.current_size) / 8
        yield RawBytes(self, "data", size)

