"""
GIF picture parser.

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, Parser,
    Enum, UInt8, UInt16,
    Bit, Bits,
    String, Character,
    PaddingBits, RawBytes)
from hachoir.error import warning, error
from hachoir.parser.image.common import Palette
from hachoir.endian import LITTLE_ENDIAN
from hachoir.text_handler import hexadecimal

class Image(FieldSet):
    def createFields(self):
        yield UInt16(self, "left", "Left")
        yield UInt16(self, "top", "Top")
        yield UInt16(self, "width", "Width")
        yield UInt16(self, "height", "Height")

        yield Bits(self, "bpp", 3, "Bits / pixel minus one")
        yield PaddingBits(self, "nul", 3, pattern=0)
        yield Bit(self, "interlaced", "Interlaced?")
        yield Bit(self, "local_map", "Use local color map?")

        if self["local_map"].value:
            nb_color = 1 << (1 + self["bpp"].value)
            yield Palette(self, "local_map", nb_color, "Local color map")

    def _createDescription(self):
        return "Image: %ux%u pixels at (%u,%u)" \
            % (self["width"].value, self["height"].value, self["left"].value, self["top"].value)

class ExtensionChunk(FieldSet):
    def createFields(self):
        yield UInt8(self, "size", "Size (in bytes)")
        size = self["size"].value
        if 0 < size:
            yield RawBytes(self, "content", size, "Content")

class Extension(FieldSet):
    def createFields(self):
        yield UInt8(self, "func", "Function", text_handler=hexadecimal)
        while True:
            field = ExtensionChunk(self, "field[]")
            yield field
            if field["size"].value == 0:
                break

    def _createDescription(self):
        return "Extension: function %s" % self["func"].display

class ScreenDescriptor(FieldSet):
    def createFields(self):
        yield UInt16(self, "width", "Width")
        yield UInt16(self, "height", "Height")
        yield Bits(self, "bpp", 3, "Bits per pixel minus one")
        yield PaddingBits(self, "nul", 1, "(reserved)", pattern=0)
        yield Bits(self, "color_res", 3, "Color resolution minus one")
        yield Bit(self, "global_map", "Has global map?")
        yield UInt8(self, "background", "Background color")
        yield UInt8(self, "notused", "Not used (zero)")

    def _createDescription(self):
        colors = 1 << (self["bpp"].value+1)
        return "Screen descriptor: %ux%u, %u colors" \
            % (self["width"].value, self["height"].value, colors)

class GifFile(Parser):
    endian = LITTLE_ENDIAN
    separator_name = {
        "!": "Extension",
        ",": "Image",
        ";": "Terminator"
    }
    tags = {
        "file_ext": "gif",
        "mime": ["image/gif"],
        "min_size": (6 + 7 + 1 + 9)*8,   # signature + screen + separator + image
        "description": "GIF picture"
    }

    def validate(self):
        if self.stream.readBytes(0, 6) not in ("GIF87a", "GIF89a"):
            return "Wrong header"
        return True

    def createFields(self):
        # Header
        yield String(self, "header", 6, "File header", charset="ascii")

        yield ScreenDescriptor(self, "screen", "Screen descriptor")
        if self["screen/global_map"].value:
            bpp = (self["screen/bpp"].value+1)
            yield Palette(self, "color_map", 1 << bpp, "Color map")
            self.color_map = self["color_map"]
        else:
            self.color_map = None

        self.images = []
        while True:
            code = Enum(Character(self, "separator[]", "Separator code"), self.separator_name)
            yield code
            code = code.value
            if code == "!":
                yield Extension(self, "extensions[]")
            elif code == ",":
                yield Image(self, "image[]")
                size = (self._size - self.current_size)/8
                if 0 < size:
                    yield RawBytes(self, "end", size)
                # TODO: Write Huffman parser code :-)
                return
            elif code == ";":
                # GIF Terminator
                break
            else:
                error("Wrong GIF image separator: ASCII %02X." % ord(code))
                break

