"""
ICO picture file format parser.

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, Parser, ParserError,
    UInt8, UInt16, UInt32, Enum, RawBytes)
from hachoir.parser.image.common import PaletteRGBA
from hachoir.endian import LITTLE_ENDIAN
from hachoir.error import warning

class BitmapInfoHeader(FieldSet):
    """ Win32 BITMAPINFOHEADER structure from GDI """
    static_size = 40*8

    compression_name = {
        0: "Uncompressed (RGB)",
        1: "RLE (8 bits)",
        2: "RLE (4 bits)",
        3: "Bitfields",
        4: "JPEG",
        5: "PNG"
    }

    def createFields(self):
        yield UInt32(self, "hdr_size", "Header size (in bytes) (=40)")
        yield UInt32(self, "width", "Width")
        yield UInt32(self, "height", "Height")
        yield UInt16(self, "nb_planes", "Color planes")
        yield UInt16(self, "bpp", "Bits/pixel")
        yield Enum(UInt32(self, "compression", "Compression"), self.compression_name)
        yield UInt32(self, "size", "Image size (in bytes)")
        yield UInt32(self, "xres", "X pixels per meter")
        yield UInt32(self, "yres", "Y pixels per meter")
        yield UInt32(self, "color_used", "Number of used colors")
        yield UInt32(self, "color_important", "Number of important colors")

    def _createDescription(self):
        return "Bitmap info header: %ux%u pixels, %u bits/pixel" % \
            (self["width"].value, self["height"].value, self["bpp"].value)

class IconHeader(FieldSet):
    def createFields(self):
        yield UInt8(self, "width", "Width")
        yield UInt8(self, "height", "Height")
        yield UInt8(self, "nb_color", "Number of colors")
        yield UInt8(self, "reserved", "(reserved)")
        yield UInt16(self, "planes", "Color planes (=1)")
        yield UInt16(self, "bpp", "Bits per pixel")
        yield UInt32(self, "size", "Content size in bytes")
        yield UInt32(self, "offset", "Data offset")

    def _createDescription(self):
        return "Icon: %ux%u pixels, %u bits/pixel" % \
            (self["width"].value, self["height"].value, self["bpp"].value)

class IconData(FieldSet):
    def __init__(self, parent, name, header, description="Icon data"):
        FieldSet.__init__(self, parent, name, description)
        self.header = header

    def createFields(self):
        yield BitmapInfoHeader(self, "header")

        # Read palette if needed
        nb_color = self.header["nb_color"].value
        if self.header["bpp"].value == 8:
            nb_color = 256
        if nb_color != 0:
            yield PaletteRGBA(self, "palette", nb_color)

        # Read pixels
        size = self.header["size"].value - self.current_size/8
        yield RawBytes(self, "pixels", size, "Image pixels")

class IcoFile(Parser):
    endian = LITTLE_ENDIAN
    tags = {
        "file_ext": "ico",
        "mime": ["image/x-ico"],
        "min_size": (22 + 40)*8,
        "description": "Microsoft icon or cursor"
    }
    type_name = {
        1: "Icon",
        2: "Mouse cursor"
    }

    def validate(self):
        # Check signature and type
        if self["signature"].value != 0:
            return "Wrong file signature"
        if self["type"].value not in self.type_name:
            return "Unknow picture type"

        # Check first header: bpp and reserved
        header = self["icon_header[0]"]
        if header["bpp"].value not in (1, 2, 4, 8, 16, 24, 32):
            return "Wrong bits/pixel value"

        # Don't check "header["reserved"].value == 0" since ImageMagick
        # generate icon with reserved=183 !?
        return True

    def createFields(self):
        yield UInt16(self, "signature", "Signature (0x0000)")
        yield Enum(UInt16(self, "type", "Resource type"), self.type_name)
        yield UInt16(self, "nb_items", "Number of items")
        items = []
        for i in range(0, self["nb_items"].value):
            item = IconHeader(self, "icon_header[]", "Icon header %u" % i)
            yield item
            items.append(item)
        for header in items:
            if header["offset"].value*8 != self.current_size:
                raise ParserError("Problem with icon data offset.")
            yield IconData(self, "icon_data[]", header)

