"""
JPEG picture parser.

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, Parser, ParserError,
    UInt8, UInt16, Enum,
    String, RawBytes)
from hachoir.parser.image.common import Palette
from hachoir.endian import BIG_ENDIAN
from hachoir.text_handler import hexadecimal
from hachoir.parser.image.exif import Exif
from hachoir.parser.image.photoshop_metadata import PhotoshopMetadata

class JpegChunkApp0(FieldSet):
    unit_name = {
        0: "pixels",
        1: "dots per inch",
        2: "dots per cm"
    }

    def createFields(self):
        yield String(self, "jfif", 5, "JFIF string", charset="ascii")
        if self["jfif"].value != "JFIF\0":
            raise ParserError(
                "Stream doesn't look like JPEG chunk (wrong JFIF signature)")
        yield UInt8(self, "ver_maj", "Major version")
        yield UInt8(self, "ver_min", "Minor version")
        yield Enum(UInt8(self, "units", "Units"), self.unit_name)
        if self["units"].value == 0:
            yield UInt16(self, "aspect_x", "Aspect ratio (X)")
            yield UInt16(self, "aspect_y", "Aspect ratio (Y)")
        else:
            yield UInt16(self, "x_density", "X density")
            yield UInt16(self, "y_density", "Y density")
        yield UInt8(self, "thumb_w", "Thumbnail width")
        yield UInt8(self, "thumb_h", "Thumbnail height")
        thumb_size = self["thumb_w"].value * self["thumb_h"].value
        if thumb_size != 0:
            yield Palette(self, "thumb_palette", 256)
            yield RawBytes(self, "thumb_data", thumb_size, "Thumbnail data")

class StartOfFrame(FieldSet):
    def createFields(self):
        yield UInt8(self, "precision")
        if self["precision"].value != 8:
            raise ParserError("JPEG error: Only 8-bit precision is supported")

        yield UInt16(self, "height")
        yield UInt16(self, "width")
        yield UInt8(self, "nr_components")
        if self["nr_components"].value not in (1, 3):
            raise ParserError("JPEG error: Only gray or true-color JPGS supported")

        for index in range(self["nr_components"].value):
            yield UInt8(self, "component_id[]")
            yield UInt8(self, "high[]")
            yield UInt8(self, "low[]")

class StartOfScan(FieldSet):
    def createFields(self):
        yield UInt8(self, "nr_components")
        if self["nr_components"].value not in (1, 3):
            raise ParserError("JPEG error: Invalid SOS marker")

        for index in range(self["nr_components"].value):
            comp_id = UInt8(self, "component_id[]")
            if (comp_id.value == 0) or (self["nr_components"].value < comp_id.value):
               raise ParserError("JPEG error: Invalid component-id")
            yield comp_id
            yield UInt8(self, "value[]")
        yield RawBytes(self, "raw", 3) # TODO: What's this???

class DQT(FieldSet):
    def createFields(self):
        while self.current_size < self._size:
            info = UInt8(self, "qtinfo[]")
            if (info.value >> 4) != 0:
                raise ParserError("JPEG error: 16 bit quantization table not supported")
            yield info
            yield RawBytes(self, "qt_table[]", 64, "8x8 quant table")

class JpegChunk(FieldSet):
    type_description = {
        0xC0: "Start Of Frame 0 (SOF0)",
        0xC4: "Define Huffman Table (DHT)",
        0xD8: "Start of image (SOI)",
        0xD9: "End of image (EOI)",
        0xDA: "Start Of Scan (SOS)",
        0xDB: "Define Quantization Table (DQT)",
        0xDC: "Define number of Lines (DNL)",
        0xDD: "Define Restart Interval (DRI)",
        0xE0: "APP0",
        0xED: "Photoshop marker",
        0xFE: "Comment"
    }
    type_name = {
        0xC0: "sof",
        0xC4: "dht[]",
        0xD8: "soi",
        0xD9: "eoi",
        0xDA: "sos",
        0xDB: "dqt[]",
        0xDC: "dnl[]",
        0xDD: "dri[]",
        0xE0: "app0",
        0xED: "psd",
        0xFE: "comment[]"
    }
    handler = {
        0xE0: JpegChunkApp0,
        0xC0: StartOfFrame,
        0xDA: StartOfScan,
        0xDB: DQT,
        0xED: PhotoshopMetadata
    }

    def __init__(self, parent, name, description=None):
        FieldSet.__init__(self, parent, name, description)
        type = self["type"].value
        if type in self.type_name:
            self._name = self.type_name[type]
        elif type == 0xE1:
            bytes = self.stream.readBytes(self.absolute_address + 32, 6)
            if bytes == "Exif\0\0":
                self._name = "exif"
                self._description = "EXIF"
#            else:
#                self._name = "adobe_metadata"
#                self._description = "Adobe Metadata XML"

    def createFields(self):
        yield UInt8(self, "header", "Header", text_handler=hexadecimal)
        if self["header"].value != 0xFF:
            raise ParserError("Invalid JPEG chunk header!")
        yield Enum(UInt8(self, "type", "Type", text_handler=hexadecimal), self.type_description)
        yield UInt16(self, "size", "Size")
        type = self["type"].value
        size = (self["size"].value - 2)
        if 0 < size:
            if type in JpegChunk.handler:
                handler = JpegChunk.handler[type]
            elif self._name == "exif":
                handler = Exif
            else:
                handler = None
            if handler:
                yield handler(self, "content", "Chunk content", size=size*8)
            else:
                yield RawBytes(self, "data", size, "Data")

    def _createDescription(self):
        return "Chunk: %s" % self["type"].display

class JpegFile(Parser):
    endian = BIG_ENDIAN
    tags = {
        "file_ext": ("jpg", "jpeg"),
        "mime": ["image/jpeg"],
        "min_size": 22*8,
        "description": "JPEG picture"
    }

    def validate(self):
        return (self["soi"].value == 0xFFD8)

    def createFields(self):
        yield UInt16(self, "soi", "\"Start of image\" (SOI) tag (0xFFD8)", text_handler=hexadecimal)
        while True:
            chunk = JpegChunk(self, "chunk[]")
            yield chunk
            if chunk["type"].value == 0xDA:
                break
        size = (self._size - self.current_size) / 8 - 2
        yield RawBytes(self, "data", size, "JPEG data")
        yield UInt16(self, "eoi", "\"End of image\" (EOI) (0xFFD9)", text_handler=hexadecimal)
        if self["eoi"].value != 0xFFD9:
            raise ParserError(
                "Stream doesn't look like JPEG file (wrong EOI tag)")

