"""
PNG picture file parser.

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, Parser, ParserError,
    UInt8, UInt16, UInt32,
    String, CString,
    Bytes, RawBytes,
    Bit, Bits, RawBits,
    Enum)
from hachoir.parser.image.common import RGB
from hachoir.text_handler import hexadecimal
from hachoir.endian import NETWORK_ENDIAN
from hachoir.error import warning
import datetime

compression_name = {
    0: "deflate" # with 32K sliding window
}

def headerParse(self):
    yield UInt32(self, "width", "Width (pixels)")
    yield UInt32(self, "height", "Height (pixels)")
    yield UInt8(self, "bpp", "Bits per pixel")
    yield RawBits(self, "reserved", 5)
    yield Bit(self, "alpha", "Alpha channel used?")
    yield Bit(self, "color", "Color used?")
    yield Bit(self, "palette", "Palette used?")
    yield Enum(UInt8(self, "compression", "Compression method"), compression_name)
    yield UInt8(self, "filter", "Filter method")
    yield UInt8(self, "interlace", "Interlace method")

def headerDescription(self):
    return "Header: %ux%u pixels and %u bits/pixel" % \
        (self["width"].value, self["height"].value, self["bpp"].value)

def paletteParse(self):
    return "Palette: %u colors" % self.nb_colors

def paletteDescription(self):
    size = self["size"].value
    if (size % 3) != 0:
        raise ParserError("Palette have invalid size (%s), should be 3*n!" % size)
    nb_color = size / 3
    for index in range(self.nb_colors):
        yield RGB(self, "color[]")

def gammaParse(self):
    yield UInt32(self, "gamma", "Gamma (x10,000)")
def gammaValue(self):
    return float(self["gamma"].value) / 10000
def gammaDescription(self):
    return "Gamma: %.3f" % self.value

def textParse(self):
    yield CString(self, "keyword", "Keyword", charset="ISO 8859-1")
    length = self["size"].value - self["keyword"].size/8
    yield String(self, "text", length, "Text", charset="ISO 8859-1")

def textDescription(self):
    return 'Text: %s' % self["text"].display

def timestampParse(self):
    yield UInt16(self, "year", "Year")
    yield UInt8(self, "month", "Month")
    yield UInt8(self, "day", "Day")
    yield UInt8(self, "hour", "Hour")
    yield UInt8(self, "minute", "Minute")
    yield UInt8(self, "second", "Second")

def timestampValue(self):
    return datetime.datetime(
        self["year"].value, self["month"].value, self["day"].value,
        self["hour"].value, self["minute"].value, self["second"].value)

unit_name = {
    0: "Unknow",
    1: "Meter"
}

def physicalParse(self):
    yield UInt32(self, "pixel_per_unit_x", "Pixel per unit, X axis")
    yield UInt32(self, "pixel_per_unit_y", "Pixel per unit, Y axis")
    yield Enum(UInt8(self, "unit", "Unit type"), unit_name)

def physicalDescription(self):
    x = self["pixel_per_unit_x"].value
    y = self["pixel_per_unit_y"].value
    desc = "Physical: %ux%u pixels" % (x,y)
    if self["unit"].value == 1:
        desc += " per meter"
    return desc

class Chunk(FieldSet):
    tag_info = {
        "tIME": ("time", timestampParse, "Timestamp", timestampValue),
        "pHYs": ("physical", physicalParse, physicalDescription, None),
        "IHDR": ("header", headerParse, headerDescription, None),
        "PLTE": ("palette", paletteParse, paletteDescription, None),
        "gAMA": ("gamma", gammaParse, gammaDescription, gammaValue),
        "tEXt": ("text[]", textParse, textDescription, None),

        "bKGD": ("background", None, "Background color", None),
        "IDAT": ("data[]", None, "Image data", None),
        "iTXt": ("utf8_text[]", None, "International text (encoded in UTF-8)", None),
        "zTXt": ("comp_text[]", None, "Compressed text", None),
        "IEND": ("end", None, "End", None)
    }

    def _createValueFunc(self):
        return self.value_func(self)

    def __init__(self, parent, name, description=None):
        FieldSet.__init__(self, parent, name, description)
        self._size = (self["size"].value + 3*4) * 8
        tag = self["tag"].value
        self.desc_func = None
        self.value_func = None
        if tag in self.tag_info:
            self._name, self.parse_func, desc, value_func = self.tag_info[tag]
            if value_func:
                self.value_func = value_func
                self._createValue = self._createValueFunc
            if desc:
                if isinstance(desc, str):
                    self._description = desc
                else:
                    self.desc_func = desc
        else:
            self._description = ""
            self.parse_func = None

    def createFields(self):
        yield UInt32(self, "size", "Size")
        yield String(self, "tag", 4, "Tag", charset="ASCII")

        size = self["size"].value
        if size != 0:
            if self.parse_func:
                for field in self.parse_func(self):
                    yield field
            else:
                yield RawBytes(self, "content", size, "Data")
        yield UInt32(self, "crc32", "CRC32", text_handler=hexadecimal)

    def _createDescription(self):
        if self.desc_func:
            return self.desc_func(self)
        else:
            return "Chunk: %s" % self["tag"].display

class PngFile(Parser):
    tags = {
        "file_ext": "png",
        "mime": ("image/png", "image/x-png"),
        "min_size": 8*8, # just the identifier
        "description": "Portable Network Graphics (PNG) picture"
    }
    endian = NETWORK_ENDIAN

    def validate(self):
        return self["id"].value == '\x89PNG\r\n\x1A\n'

    def createFields(self):
        yield Bytes(self, "id", 8, r"PNG identifier ('\x89PNG\r\n\x1A\n')")
        while self.current_size < self.size:
            yield Chunk(self, "chunk[]")

