"""
Python compiled source code parser.

Informations:
 - Python/marshal.c in Python 2.4.2 source code

Author: Victor Stinner
Creation: 25 march 2005
"""

from hachoir.field import (Parser, FieldSet,
    UInt16, UInt32, UInt64, Enum,
    Character, Bytes, RawBytes)
from hachoir.endian import LITTLE_ENDIAN
from hachoir.text_handler import hexadecimal
from hachoir.error import error

class PythonString(FieldSet):
    def createFields(self):
        yield UInt32(self, "length", "Length")
        length = self["length"].value
        if 0 < length:
            yield RawBytes(self, "content", length, "Content")

    def _createDescription(self):
        return "String (len=%s)" % self["length"].value

class Interned(PythonString):
    def _createDescription(self):
        return "Interned (len=%s)" % self["length"].value

class Unicode(PythonString):
    def _createDescription(self):
        return "Unicode (len=%s)" % self["length"].value

class Float(FieldSet):
    def createFields(self):
        yield PascalString8(self, "content", "Content")

    def _createDescription(self):
        return "Float: %s" % self["content"].value

class Complex(FieldSet):
    def createFields(self):
        yield PascalString8(self, "real_part")
        yield PascalString8(self, "complex_part")

    def _createDescription(self):
        return "Complex: %s+%sb" % (self["real_content"].value, self["complex_content"].value)

class Tuple(FieldSet):
    def createFields(self):
        yield UInt32(self, "count", "Item count")
        for index in xrange(self["count"].value):
            yield Object(self, "item[]")

    def _createDescription(self):
        return "Tuple: %s item(s)" % self["count"].value

class List(Tuple):
    def _createDescription(self):
        return "List: %s item(s)" % self["count"].value

class Dict(FieldSet):
    def createFields(self):
        self.count = 0
        while True:
            key = Object(self, "key[]")
            yield key
            if key["bytecode"].value == "0":
                break
            yield Object(self, "value[]")
            self.count += 1

    def _getDescription(self):
        return "Dict: %s key(s)" % self.count
    description = property(_getDescription)

class Long(FieldSet):
    def createFields(self):
        yield UInt32(self, "digit_count")
        for i in xrange( self["digit_count"].value ):
            yield UInt16(self, "digit[]")

    def _createDescription(self):
        return "Long: %s digit(s)" % self["digit_count"].value

class Integer32(FieldSet):
    def createFields(self):
        yield UInt32(self, "value")

    def _createDescription(self):
        return "Integer32: %s" % self["value"].value

class Integer64(FieldSet):
    def createFields(self):
        yield UInt64(self, "value")

    def _createDescription(self):
        return "Integer64: %s" % self["value"].value

class StringRef(FieldSet):
    def createFields(self):
        yield UInt32(self, "ref", text_handler=hexadecimal)

    def _createDescription(self):
        return "String ref.: %s" % self["ref"].display

class Code(FieldSet):
    def createFields(self):
        yield UInt32(self, "arg_count", "Argument count")
        yield UInt32(self, "nb_locals", "Number of local variables")
        yield UInt32(self, "stack_size", "Stack size")
        yield UInt32(self, "flags")
        yield Object(self, "code")
        yield Object(self, "consts")
        yield Object(self, "names")
        yield Object(self, "varnames")
        yield Object(self, "freevars")
        yield Object(self, "cellvars")
        yield Object(self, "filename")
        yield Object(self, "name")
        yield UInt32(self, "firstlineno", "First line number")
        yield Object(self, "lnotab")

class Object(FieldSet):
    bytecode_name = {
        '0': "NULL",
        'N': "None",
        'F': "False",
        'T': "True",
        "S": "StopIter",
        ".": "ELLIPSIS",

        "i": "Int",
        "I": "Int64",
        "f": "Float",
        "x": "Complex",
        "l": "Long",
        "s": "String",
        "t": "Interned",
        "R": "StringRef",
        "(": "Tuple",
        "[": "List",
        "{" : "Dict",
        'c': "Code",
        "u": "Unicode",
        "?": "Unknow"
    }
    bytecode_handler =  {
        'c': Code,
        's': PythonString,
        'u': Unicode,
        "R": StringRef,
        '(': Tuple,
        '[': List,
        "i": Integer32,
        "I": Integer64,
        "t": Interned,
        "f": Float,
        "x": Complex,
        "l": Long,
        "{": Dict,

        # Don't contains any data
        "T": None,
        'F': None,
        'S': None,
        '.': None,
        '0': None,
        'N': None
    }

    def createFields(self):
        yield Enum(Character(self, "bytecode", "Bytecode"), Object.bytecode_name)
        bytecode = self["bytecode"].value
        if bytecode in Object.bytecode_handler:
            cls = Object.bytecode_handler[bytecode]
            if cls is not None:
                yield cls(self, "content")
        else:
            error("Compiled Python parser: Don't have parser for bytecode \"%s\"." % bytecode)
            raise StopIteration()

    def _createDescription(self):
        if "content" in self:
            desc = self["content"].description
        else:
            desc = self["bytecode"].display
        return "Object %s" % desc

class PythonCompiledFile(Parser):
    tags = {
        "file_ext": ("pyc", "pyo"),
        "min_size": 9*8,
        "description": "Compiled Python script"
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        return (self["signature"].value == 'm\xF2\r\n')

    def createFields(self):
        yield Bytes(self, "signature", 4, r"Python file signature (m\xF2\r\n)")
        yield RawBytes(self, "xxx", 4)
        while self.current_size < self.size:
            try:
                yield Object(self, "object[]")
                break
            except StopIteration:
                error("Compiled Python parser: STOP!")
                return

