"""
ASF (multimedia container) parser.

http://www.microsoft.com/windows/windowsmedia/forpros/format/asfspec.aspx
http://swpat.ffii.org/pikta/xrani/asf/index.fr.html

Author: Victor Stinner
Creation: 5 august 2006
"""

from hachoir.field import (Parser, FieldSet,
    UInt16, UInt32, UInt64, GenericInteger,
    String, PascalString16, Enum,
    Bit, Bits, PaddingBits,
    PaddingBytes, RawBytes)
from hachoir.endian import LITTLE_ENDIAN
from hachoir.text_handler import (hexadecimal,
    humanFilesize, timestampWin64, durationWin64, humanBitSize)
from hachoir.bits import str2hex, swap16, swap32
from hachoir.tools import paddingSize

class GUID(FieldSet):
    def createFields(self):
        # TODO: Maybe only use UInt32 or UInt16?
        yield UInt32(self, "a", text_handler=hexadecimal)
        yield UInt32(self, "b", text_handler=hexadecimal)
        yield UInt32(self, "c", text_handler=hexadecimal)
        yield UInt32(self, "d", text_handler=hexadecimal)

    def _createValue(self):
        return self.stream.readBytes(self.absolute_address, 16)

    def _getDisplay(self):
        return str2hex(self.value, format=r"\x%02X")

    def _createDescription(self):
        b = self["b"].value
        c = self["c"].value
        return "GUID=%08X-%04X-%04X-%04X-%04X%08X" % (
            self["a"].value,
            b & 0xFFFF, b >> 16,
            swap16(c & 0xFFFF), swap16(c >> 16),
            swap32(self["d"].value))

class FileProperty(FieldSet):
    """
    GUID: 8CABDCA1-A947-11CF-8EE4-00C00C205365
    """
    def createFields(self):
        yield GUID(self, "guid")
        yield UInt64(self, "file_size", text_handler=humanFilesize)
        yield UInt64(self, "creation_date", text_handler=timestampWin64)
        yield UInt64(self, "pckt_count")
        yield UInt64(self, "play_duration", text_handler=durationWin64)
        yield UInt64(self, "send_duration", text_handler=durationWin64)
        yield UInt64(self, "preroll")
        yield Bit(self, "broadcast", "Is broadcast?")
        yield Bit(self, "seekable", "Seekable stream?")
        yield PaddingBits(self, "reserved[]", 30)
        yield UInt32(self, "min_pckt_size")
        yield UInt32(self, "max_pckt_size")
        yield UInt32(self, "max_bitrate")

class HeaderExtension(FieldSet):
    """
    GUID: 5FBF03B5-A92E-11CF-8EE3-00C00C205365
    """
    def createFields(self):
        yield GUID(self, "reserved[]")
        yield UInt16(self, "reserved[]")
        yield UInt32(self, "size")
        yield PaddingBytes(self, "data", self["size"].value)

class Header(FieldSet):
    """
    GUID: 75B22630-668E-11CF-A6D9-00AA0062CE6C
    """
    def createFields(self):
        yield UInt32(self, "obj_count")
        yield PaddingBytes(self, "reserved[]", 2)
        for index in xrange(self["obj_count"].value):
            yield Object(self, "object[]")

class Descriptor(FieldSet):
    """
    See ExtendedContentDescription class.
    """
    type_name = {
        0: "Unicode",
        1: "Byte array",
        2: "BOOL (32 bits)",
        3: "DWORD (32 bits)",
        4: "QWORD (64 bits)",
        5: "WORD (16 bits)"
    }
    def createFields(self):
        yield PascalString16(self, "name", "Name", charset="UTF-16LE")
        yield Enum(UInt16(self, "type"), self.type_name)
        yield UInt16(self, "value_length")
        type = self["type"].value
        size = self["value_length"].value
        name = "value"
        if type == 0 and (size % 2) == 0:
            yield String(self, name, size/2, charset="UTF-16LE")
        elif type in (2, 3):
            yield UInt32(self, name)
        elif type == 4:
            yield UInt64(self, name)
        else:
            yield RawBytes(self, name, size)

class ExtendedContentDescription(FieldSet):
    """
    GUID: D2D0A440-E307-11D2-97F0-00A0C95EA850
    """
    def createFields(self):
        yield UInt16(self, "count")
        for index in xrange(self["count"].value):
            yield Descriptor(self, "descriptor[]")

class Codec(FieldSet):
    """
    See CodecList class.
    """
    type_name = {
        1: "video",
        2: "audio"
    }
    def createFields(self):
        yield Enum(UInt16(self, "type"), self.type_name)
        yield UInt16(self, "name_len")
        yield String(self, "name", self["name_len"].value, "Name", charset="UTF-16LE")
        yield UInt16(self, "desc_len")
        yield String(self, "desc", self["desc_len"].value, "Description", charset="UTF-16LE")
        yield UInt16(self, "info_len")
        size = self["info_len"].value
        if size:
            yield RawBytes(self, "info", size)

class CodecList(FieldSet):
    """
    GUID: 86D15240-311D-11D0-A3A4-00A0C90348F6
    """
    def createFields(self):
        yield GUID(self, "reserved[]")
        yield UInt32(self, "count")
        for index in xrange(self["count"].value):
            yield Codec(self, "codec[]")

class SimpleIndexEntry(FieldSet):
    """
    See SimpleIndex class.
    """
    def createFields(self):
        yield UInt32(self, "pckt_number")
        yield UInt16(self, "pckt_count")

class SimpleIndex(FieldSet):
    """
    GUID:
    """
    def createFields(self):
        yield GUID(self, "file_id")
        yield UInt64(self, "entry_interval", text_handler=durationWin64)
        yield UInt32(self, "max_pckt_count")
        yield UInt32(self, "entry_count")
        for index in xrange(self["entry_count"].value):
            yield SimpleIndexEntry(self, "entry[]")

class BitRate(FieldSet):
    """
    See BitRateList class.
    """
    def createFields(self):
        yield Bits(self, "stream_index", 7)
        yield PaddingBits(self, "reserved", 9)
        yield UInt32(self, "avg_bitrate", text_handler=humanBitSize)

class BitRateList(FieldSet):
    """
    GUID: 7BF875CE-468D-11D1-8D82-006097C9A2B2
    """
    def createFields(self):
        yield UInt16(self, "count")
        for index in xrange(self["count"].value):
            yield BitRate(self, "bit_rate[]")

class Data(FieldSet):
    """
    GUID: 75B22636-668E-11CF-A6D9-00AA0062CE6C
    """
    def createFields(self):
        yield GUID(self, "file_id")
        yield UInt64(self, "packet_count")
        yield PaddingBytes(self, "reserved", 2)
        size = (self.size - self.current_size) / 8
        yield RawBytes(self, "data", size)

class StreamProperty(FieldSet):
    """
    GUID: B7DC0791-A9B7-11CF-8EE6-00C00C205365
    """
    def createFields(self):
        yield GUID(self, "type")
        yield GUID(self, "error_correction")
        yield UInt64(self, "time_offset")
        yield UInt32(self, "data_len")
        yield UInt32(self, "error_correct_len")
        yield Bits(self, "stream_index", 7)
        yield Bits(self, "reserved[]", 8)
        yield Bit(self, "encrypted", "Content is encrypted?")
        yield UInt32(self, "reserved[]")
        size = self["data_len"].value
        if size:
            yield RawBytes(self, "data", size)
        size = self["error_correct_len"].value
        if size:
            yield RawBytes(self, "error_correct", size)

class Object(FieldSet):
    tag_info = {
        "\x30\x26\xB2\x75\x8E\x66\xCF\x11\xA6\xD9\x00\xAA\x00\x62\xCE\x6C":
            ("header", Header, "Header object"),
        "\xA1\xDC\xAB\x8C\x47\xA9\xCF\x11\x8E\xE4\x00\xC0\x0C\x20\x53\x65":
            ("file_prop", FileProperty, "File property"),
        "\xB5\x03\xBF\x5F\x2E\xA9\xCF\x11\x8E\xE3\x00\xC0\x0C\x20\x53\x65":
            ("header_ext", HeaderExtension, "Header extension"),
        "\x40\x52\xD1\x86\x1D\x31\xD0\x11\xA3\xA4\x00\xA0\xC9\x03\x48\xF6":
            ("codec_list", CodecList, "Codec list"),
        "\x90\x08\x00\x33\xB1\xE5\xCF\x11\x89\xF4\x00\xA0\xC9\x03\x49\xCB":
            ("simple_index", SimpleIndex, "Simple index"),
        "\x36\x26\xB2\x75\x8E\x66\xCF\x11\xA6\xD9\x00\xAA\x00\x62\xCE\x6C":
            ("data", Data, "Data object"),
        "\x91\x07\xDC\xB7\xB7\xA9\xCF\x11\x8E\xE6\x00\xC0\x0C\x20\x53\x65":
            ("stream_prop[]", StreamProperty, "Stream properties"),
        "\xCE\x75\xF8\x7B\x8D\x46\xD1\x11\x8D\x82\x00\x60\x97\xC9\xA2\xB2":
            ("bit_rates[]", BitRateList, "Bit rate list"),
        "\x40\xA4\xD0\xD2\x07\xE3\xD2\x11\x97\xF0\x00\xA0\xC9\x5E\xA8\x50":
            ("ext_content_desc", ExtendedContentDescription,
             "Extended content description"),
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)

        tag = self["guid"].value
        if tag not in self.tag_info:
            self.handler = None
            return
        info = self.tag_info[tag]
        self._name = info[0]
        self.handler = info[1]

    def createFields(self):
        yield GUID(self, "guid")
        yield UInt64(self, "size")

        size = self["size"].value - self.current_size/8
        if self.handler:
            yield self.handler(self, "content", size=size*8)
        else:
            yield RawBytes(self, "content", size)

class AsfFile(Parser):
    tags = {
        "file_ext": ("wmv", "wma", "asf"),
        "mime": ("video/x-ms-asf", "video/x-ms-wmv", "audio/x-ms-wma"),
        "min_size": 24*8,
        "description": "ASF container (WMV video, WMA audio)"
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        guid = self.stream.readBytes(0, 16)
        # TODO: Set right MIME type
#        if (...):
#            self._mime_type = "video/x-ms-wmv"
#        elif (...):
#            self._mime_type = "audio/x-ms-wma"
        return guid == "\x30\x26\xB2\x75\x8E\x66\xCF\x11\xA6\xD9\x00\xAA\x00\x62\xCE\x6C"

    def createFields(self):
        while self.current_size < self.size:
            yield Object(self, "object[]")

