"""
Parser list managment:
- loadParserPlugins() load all parsers ;
- getParsersByStream() find right plugin from a binary stream.
"""

import os
import types
from hachoir.error import error, warning, info, HachoirError
from hachoir.field import Parser
from hachoir import config

_hachoir_parsers_list = None

def loadParserPlugins():
    """
    Load all parsers from "hachoir.parser" module.

    Return a list of loaded plugins.
    """
    global _hachoir_parsers_list

    # Parser list is already loaded?
    if _hachoir_parsers_list:
        return _hachoir_parsers_list

    _hachoir_parsers_list = []
    todo = []
    module = __import__("hachoir.parser").parser
    for attrname in dir(module):
        attr = getattr(module, attrname)
        if isinstance(attr, types.ModuleType):
            todo.append(attr)

    for module in todo:
        for attrname in dir(module):
            attr = getattr(module, attrname)
            if isinstance(attr, type):
                _hachoir_parsers_list.append(attr)
    assert 1 <= len(_hachoir_parsers_list)
    return _hachoir_parsers_list

def printParserList():
    print "List of Hachoir parsers."

    # Get parser list
    parser_list = loadParserPlugins()

    # Create parser list sorted by module
    parsers_by_module = {}
    for parser in parser_list:
        module = parser.__module__.split(".")[2]
        if module in parsers_by_module:
            parsers_by_module[module].append(parser)
        else:
            parsers_by_module[module] = [parser]
    for module in sorted(parsers_by_module):
        print "\n[%s]" % module
        for parser in parsers_by_module[module]:
            text = ["- "]
            mimes = parser.tags.get("mime", ())
            if not isinstance(mimes, str):
                if len(mimes):
                    mimes = ", ".join(mimes)
                    text.append(mimes)
                    text.append(": ")
                else:
                    mimes = "(no mime type)"
            text.append(parser.tags["description"])
            print "".join(text)
    print

################################################################################

def _createParser(cls, stream, verbose=False):
    try:
        if config.debug:
            info("Guess %s" % cls.__name__)
        parser = cls(stream)
        result = parser.validate()
        if result is True:
            return parser
        if isinstance(result, str):
            if verbose:
                warning("Skip parser %s: %s" % (cls.__name__, result))
    except HachoirError, err:
        error("Unable to create parser %s: %s" % (cls.__name__, err))
    return None

def _matchTag(valid, file_ext):
    if isinstance(valid, list):
        return file_ext in valid
    elif isinstance(valid, tuple):
        for item in valid:
            if file_ext == item:
                return True
        return False
    else:
        return valid == file_ext

def _guessParserByTag(stream, parser_list, tag_name, tag_value):
    next_try = []
    for cls in parser_list:
        if tag_name in cls.tags \
        and _matchTag(cls.tags[tag_name], tag_value):
            parser = _createParser(cls, stream, True)
            if parser:
                return True, parser
        else:
            next_try.append(cls)
    return False, next_try

def _guessParsers(stream, file_ext=None, mime=None):
    size = stream.size
    parser_list = loadParserPlugins()

    # Filter by minimum size
    parser_list = [ parser for parser in parser_list \
        if "min_size" not in parser.tags or parser.tags["min_size"] <= size ]

    # Guess my MIME type
    if mime:
        ok, result = _guessParserByTag(stream, parser_list, "mime", mime)
        if ok:
            return result
        else:
            parser_list = result

    # Guess by file extension
    if file_ext:
        ok, result = _guessParserByTag(stream, parser_list, "file_ext", file_ext)
        if ok:
            return result
        else:
            parser_list = result

    # Guess other parser
    for cls in parser_list:
       parser = _createParser(cls, stream)
       if parser:
           return parser
    return None

def guessParser(stream, filename=None, mime=None):
    # Extract file extension
    filename = os.path.basename(filename)
    filename = filename.split(".")

    # Guess parser
    if 1 < len(filename):
        return _guessParsers(stream, file_ext=filename[-1], mime=mime)
    else:
        return _guessParsers(stream, mime=mime)

