from cStringIO import StringIO
from hachoir.endian import BIG_ENDIAN, LITTLE_ENDIAN
from hachoir.error import HachoirError
from hachoir.bits import str2long
from hachoir.tools import lower_bound

class InputStreamError(HachoirError):
    pass


class InputStream:
    def __init__(self, source=None, size=None):
        self.source = source
        self._size = size
        if size == 0:
            raise InputStreamError(
                "Error: input size is nul (source='%s')!" % source)

    def _getSize(self):
        return self._size
    size = property(_getSize, doc="Size of the stream in bits")

    def read(self, address, size):
        """
        Read 'size' bits at position 'address' (in bits)
        from the beginning of the stream.
        """
        raise NotImplementedError

    def readBits(self, address, nbits, endian):
        assert endian in (BIG_ENDIAN, LITTLE_ENDIAN)

        shift, data = self.read(address, nbits)
        value = str2long(data, endian)
        if endian is BIG_ENDIAN:
            value >>= len(data) * 8 - shift - nbits
        else:
            value >>= shift
        return value & (1 << nbits) - 1

    def readInteger(self, address, signed, nbits, endian):
        """ Read an integer number """
        value = self.readBits(address, nbits, endian)

        # Signe number. Example with nbits=8:
        # if 128 <= value: value -= 256
        if signed and (1 << (nbits-1)) <= value:
            value -= (1 << nbits)
        return value

    def readBytes(self, address, nb_bytes):
        """ (don't worry about endianness """
        shift, data = self.read(address, 8 * nb_bytes)
        assert not shift # TODO :D
        return data

    def searchBytesLength(self, needle, include_needle,
    start_address=0, end_address=None):
        """
        If include_needle is True, add its length to the result
        """

        if end_address is None:
            end_address = (self.size / 8 - 1) * 8
        if end_address < start_address:
            return None

        pos = self.searchBytes(needle, start_address, end_address)
        if pos is None:
            return None
        length = (pos - start_address) / 8
        if include_needle:
            length += len(needle)
        return length

    def searchBytes(self, needle, start_address=0, end_address=None):
        """
        Search some bytes between start_address and end_address

        Returns the address of the bytes if founded, None else
        """

        if end_address is None:
            end_address = (self.size / 8 - 1) * 8
        if end_address < start_address:
            return None

        # TODO: Doesn't suppport bit address yet :-(
        assert (start_address % 8) == 0
        assert (end_address % 8) == 0

        size = 2048
        if size <= len(needle):
            raise InputStreamError("Search string too big.")
        doublesize = size * 2

        address = start_address
        maxsize = (end_address - start_address)/8 + 1
        if maxsize <= 0:
            return None
        if maxsize < doublesize:
            doublesize = maxsize
            size = 0
        data = self.readBytes(address, doublesize)

        new_address = address + size
        while len(data) != 0:
            found = data.find(needle)
            if found != -1:
                return address + found*8
            address = new_address
            if end_address < address + size:
                size = end_address - address
            if size == 0:
                break
            data = data[size:] + self.readBytes(address, size)
            new_address = address + size
        return None


class InputIOStream(InputStream):
    def __init__(self, input, source=None, size=None):
        self._input = input
        InputStream.__init__(self, source, size)

    def read(self, address, size):
        assert size > 0
        address, shift = divmod(address, 8)
        self._input.seek(address)
        size = (size + shift + 7) / 8
        data = self._input.read(size)
        if len(data) != size:
            raise InputStreamError(
                "Can't read %u bytes at address %u (got %u bytes)!"
                % (size, address, len(data)))
        return shift, data


def StringInputStream(content):
    inputio = StringIO(content)
    return InputIOStream(inputio, size=len(content)*8)


def FileInputStream(filename):
    try:
        inputio = open(filename, 'rb')
    except IOError, err:
        raise InputStreamError("Unable to open file (%s): %s"
            % (filename, err))
    inputio.seek(0, 2)
    size = inputio.tell() * 8
    inputio.seek(0)
    if size == 0:
        raise InputStreamError("File is empty (%s)!" % filename)
    return InputIOStream(inputio, 'file://'+filename, size)


class InputFieldStream(InputStream):
    def __init__(self, field):
        self.stream = field.parent.stream
        self.address = field.absolute_address
        InputStream.__init__(self, self.stream.source + field.path, field.size)

    def read(self, address, size):
        return self.stream.read(self.address + address, size)


class FragmentedStream(InputStream):
    def __init__(self, field):
        self.stream = field.parent.stream
        self.address = field.absolute_address
        source = self.stream.source + field.path
        self.fragments = []
        address = 0
        while field:
            data = field["data"]
            self.fragments += [ (address, data) ]
            address += data.size
            field = field.next
        InputStream.__init__(self, source, address)

    def read(self, address, size):
        d = []
        i = lower_bound(self.fragments, lambda x: x[0] <= address)
        a, f = self.fragments[i-1]
        o = address - a
        s = None
        while True:
            n = min(f.size - o, size)
            u, v = self.stream.read(f.absolute_address + o, n)
            if s is None:
                s = u
            else:
                assert not u
            d += [ v ]
            size -= n
            if not size:
                return s, ''.join(d)
            a, f = self.fragments[i]
            i += 1
            o = 0
