"""
Utilities used to convert a chunk to human classic reprentation of data.
"""

import datetime
from hachoir.tools import (
    humanDuration as doHumanDuration,
    humanFilesize as doHumanFilesize,
    humanBitSize as doHumanBitSize,
    humanFrequency as doHumanFrequency)
from hachoir.bits import str2bin

def timestampWin64(chunk):
    """
    Convert Windows 64-bit timestamp to string. The timestamp format is
    a 64-bit number which represents number of 100ns since the
    1st January 1601.

    See also durationWin64()
    """
    assert chunk.size == 64
    if 0 < chunk.value:
        date  = datetime.datetime(1601, 1, 1, 0, 0, 0)
        date += datetime.timedelta(microseconds=chunk.value/10)
        return str(date)
    else:
        return "(not set)"

def durationWin64(field):
    """
    Convert Windows 64-bit duration to string. The timestamp format is
    a 64-bit number which represents number of 100ns since the
    1st January 1601.

    See also timestampWin64()
    """
    assert field.size == 64
    return doHumanDuration(field.value/10000)

def timestampMSDOS(chunk):
    """
    Convert MS-DOS timestamp (32-bit integer) to string
    """
    assert chunk.size == 32
    val = chunk.value
    sec = 2 * (val & 31)              # 5 bits: second
    minute = (val >> 5) & 63          # 6 bits: minute
    hour = (val >> 11) & 31           # 5 bits: hour
    day = (val >> 16) & 31            # 5 bits: day of the month
    month = (val >> 21) & 15          # 4 bits: month
    year = 1980 + ((val >> 25) & 127) # 7 bits: year
    try:
        return str(datetime.datetime(year, month, day, hour, minute, sec))
    except ValueError: # Wrong date
        return "invalid msdos datetime (%s)" % val

def humanFilesize(chunk):
    """
    Convert a file size to human representation (just call
    hachoir.tools.humanFilesize())
    """
    return doHumanFilesize(chunk.value)

def humanFrequency(field):
    """
    Convert a file size to human representation (just call
    hachoir.tools.humanHertz())
    """
    return doHumanFrequency(field.value)

def humanDuration(field):
    return doHumanDuration(field.value)

def humanBitSize(field):
    """
    Convert a size in bits to human representation (just call
    hachoir.tools.humanBitSize())
    """
    return doHumanBitSize(field.value)

def timestampUNIX(field):
    """
    Convert an UNIX (32-bit) timestamp to string. The format is the number
    of seconds since the 1st January 1970.
    """
    timestamp = datetime.datetime.fromtimestamp(field.value)
    return str(timestamp)

def timestampMac(field):
    """
    Convert an Mac (32-bit) timestamp to string. The format is the number
    of seconds since the 1st January 1904 (to 2040).
    """
    date  = datetime.datetime(1904, 1, 1, 0, 0, 0)
    date += datetime.timedelta(seconds=field.value)
    return str(date)

def hexadecimal(chunk):
    """
    Convert an integer to hexadecimal (in lower case).
    Example: a 16-bit chunk with value 412 becomes "0x019c"
    """
    size = chunk.size
    assert 0 < size <= 64 and not size % 8
    pattern = "0x%0" + str(size/4) + "X"
    return pattern % chunk.value

