# -*- coding: UTF-8 -*-

"""
Various utilities.
"""

import re
import stat
import sys
import traceback

def paddingSize(value, align):
    """
    Compute size of a padding field.

    >>> paddingSize(31, 4)
    1
    >>> paddingSize(32, 4)
    0
    >>> paddingSize(33, 4)
    3

    Note: (value + paddingSize(value, align)) == alignValue(value, align)
    """
    if value % align != 0:
        return align - (value % align)
    else:
        return 0

def alignValue(value, align):
    """
    Align a value to next 'align' multiple.

    >>> alignValue(31, 4)
    32
    >>> alignValue(32, 4)
    32
    >>> alignValue(33, 4)
    36

    Note: alignValue(value, align) == (value + paddingSize(value, align))
    """

    if value % align != 0:
        return value + align - (value % align)
    else:
        return value

def humanDurationNanosec(nsec):
    # Nano second
    if nsec < 1000:
        return "%u nsec" % nsec

    # Micro seconds
    usec, nsec = divmod(nsec, 1000)
    if usec < 1000:
        return "%.2f usec" % (usec+float(nsec)/1000)

    # Milli seconds
    msec, usec = divmod(usec, 1000)
    if msec < 1000:
        return "%.2f ms" % (msec + float(usec)/1000)
    return humanDuration(msec)

def humanDuration(millisec):
    """
    Convert a duration in millisecond to human natural representation.

    >>> humanDuration(213)
    '213 ms'
    >>> humanDuration(4213)
    '4.21 sec'
    >>> humanDuration(6402309)
    '1 hour(s) 46 min 42 sec'
    """

    # Milliseconds
    if millisec < 1000:
        return "%u ms" % millisec

    # Seconds
    second, millisec = divmod(millisec, 1000)
    if second < 60:
        return  "%.2f sec" % (second+float(millisec)/1000)

    # Minutes
    minute, second = divmod(second, 60)
    if minute < 60:
        return "%u min %u sec" % (minute, second)

    # Hours
    hour, minute = divmod(minute, 60)
    if hour < 24:
        if 0 < second:
            return "%u hour(s) %u min %u sec" % (hour, minute, second)
        else:
            return "%u hour(s) %u min" % (hour, minute)

    # Days
    day, hour = divmod(hour, 24)
    if day < 365:
        return "%u day(s) %u hour(s)" % (day, hour)

    # Years
    year, day = divmod(day, 365)
    if hour != 0:
        text = "%u year(s) %u day(s)" % (year, day)
    else:
        text = "%u year(s)" % (year)
    return text

def humanFilesize(size):
    """
    Convert a file size in byte to human natural representation.
    It uses the values: 1 KB is 1024 bytes, 1 MB is 1024 KB, etc.

    >>> humanFilesize(790)
    '790 bytes'
    >>> humanFilesize(256960)
    '250.9 KB'
    """
    divisor = 1024
    if size < 1000:
        return "%u bytes" % size
    units = ["KB", "MB", "GB", "TB"]
    size = float(size)
    for unit in units:
        size = size / divisor
        if size < divisor:
            return "%.1f %s" % (size, unit)
    return "%u %s" % (size, unit)

def humanBitSize(size):
    """
    Convert a size in bit to human classic representation.
    It uses the values: 1 Kbit is 1000 bits, 1 Mbit is 1000 Kbit, etc.

    >>> humanBitSize(790)
    '790 bits'
    >>> humanBitSize(256960)
    '257.0 Kbit'
    """
    divisor = 1000
    if size < divisor:
        return "%u bits" % size
    units = ["Kbit", "Mbit", "Gbit", "Tbit"]
    size = float(size)
    for unit in units:
        size = size / divisor
        if size < divisor:
            return "%.1f %s" % (size, unit)
    return "%u %s" % (size, unit)

def humanFrequency(hertz):
    """
    Convert a frequency in hertz to human classic representation.
    It uses the values: 1 KHz is 1000 Hz, 1 MHz is 1000 KMhz, etc.

    Examples:
    - humanFrequency(790) => "790 Hz"
    - humanFrequency(629469) => "629.5 KHz"
    """
    divisor = 1000
    if hertz < divisor:
        return "%u Hz" % hertz
    units = ["KHz", "MHz", "GHz", "THz"]
    hertz = float(hertz)
    for unit in units:
        hertz = hertz / divisor
        if hertz < divisor:
            return "%.1f %s" % (hertz, unit)
    return "%u %s" % (hertz, unit)

regex_control_code = re.compile("([\0-\37])")
controlchars = tuple({
        # Don't use "\0", because "\0"+"0"+"1" = "\001" = "\1" (1 character)
        # Same rease to not use octal syntax ("\1")
        ord("\n"): r"\n",
        ord("\r"): r"\r",
        ord("\t"): r"\t",
        ord("\a"): r"\a",
        ord("\b"): r"\b",
    }.get(code, '\\x%02x' % code)
    for code in xrange(32)
)

def makePrintable(data, charset, quote=False, to_unicode=False):
    r"""
    Prepare a string to make it printable in the specified charset.
    It escappe control caracters. Caracter with code bigger than 127
    are escapped if: data type is 'str' or if charset is "ASCII".

    Examples with Unicode:
    >>> aged = unicode("âgé", "UTF-8")
    >>> repr(aged)  # text type is 'unicode'
    "u'\\xe2g\\xe9'"
    >>> makePrintable("abc\0", "UTF-8")
    'abc\\x00'
    >>> makePrintable(aged, "latin1")
    '\xe2g\xe9'
    >>> makePrintable(aged, "latin1", True)
    '"\xe2g\xe9"'

    Examples with string encoded in latin1:
    >>> aged_latin = unicode("âgé", "UTF-8").encode("latin1")
    >>> repr(aged_latin)  # text type is 'str'
    "'\\xe2g\\xe9'"
    >>> makePrintable(aged_latin, "latin1")
    '\\xe2g\\xe9'
    >>> makePrintable("", "latin1")
    ''
    >>> makePrintable("", "latin1", True)
    '(empty)'

    Control codes:
    >>> makePrintable("\0\x03\x0a\x10", "latin1")
    '\\x00\\x03\\n\\x10'
    """

    if 0 < len(data):
        if not isinstance(data, unicode):
            data = unicode(data, "latin1")
            charset = "ascii"
        data = regex_control_code.sub(
            lambda regs: controlchars[ord(regs.group(1))], data)
        if quote:
            data = '"%s"' % data
    elif quote:
        data = "(empty)"
    data = data.encode(charset, "backslashreplace")
    if to_unicode:
        data = unicode(data, charset)
    return data

def getBacktrace():
    """
    Try to get backtrace as string.
    Returns "Error while trying to get backtrace" on failure.
    """
    try:
        info = sys.exc_info()
        trace = traceback.format_exception(*info)
        sys.exc_clear()
        if trace[0] != "None\n":
            return "".join(trace)
        else:
            return "Empty backtrace."
    except:
        return "Error while trying to get backtrace"

def bsearch(seq, cmp_func):
    """
    Search a value in a sequence using binary search. Returns index of the
    value, or None if the value doesn't exist.

    'seq' have to be sorted in ascending order according to the
    comparaison function ;

    'cmp_func', prototype func(x), is the compare function:
    - Return strictly positive value if we have to search forward ;
    - Return strictly negative value if we have to search backward ;
    - Otherwise (zero) we got the value.

    >>> # Search number 5 (search forward)
    ... bsearch([0, 4, 5, 10], lambda x: 5-x)
    2
    >>> # Backward search
    ... bsearch([10, 5, 4, 0], lambda x: x-5)
    1
    """
    l = 0
    u = len(seq)
    while l < u:
        i = (l + u) / 2
        d = cmp_func(seq[i])
        if d < 0:
            u = i
        elif d > 0:
            l = i + 1
        else:
            return i
    return None

def lower_bound(seq, cmp_func):
    f = 0
    l = len(seq)
    while l > 0:
        h = l // 2
        m = f + h
        if cmp_func(seq[m]):
            f = m
            f += 1
            l -= h + 1
        else:
            l = h
    return f

def humanUnixAttributes(mode):
    """
    Convert a Unix file attributes (or "file mode") to a string.

    Original source code:
    http://cvs.savannah.gnu.org/viewcvs/coreutils/lib/filemode.c?root=coreutils

    >>> humanUnixAttributes(0644)
    '-rw-r--r-- (644)'
    >>> humanUnixAttributes(02755)
    '-rwxr-sr-x (2755)'
    """

    def ftypelet(mode):
        if stat.S_ISREG (mode) or not stat.S_IFMT(mode):
            return '-'
        if stat.S_ISBLK (mode): return 'b'
        if stat.S_ISCHR (mode): return 'c'
        if stat.S_ISDIR (mode): return 'd'
        if stat.S_ISFIFO(mode): return 'p'
        if stat.S_ISLNK (mode): return 'l'
        if stat.S_ISSOCK(mode): return 's'
        return '?'

    chars = [ ftypelet(mode), 'r', 'w', 'x', 'r', 'w', 'x', 'r', 'w', 'x' ]
    for i in xrange(1, 10):
        if not mode & 1 << 9 - i:
            chars[i] = '-'
    if mode & stat.S_ISUID:
        if chars[3] != 'x':
            chars[3] = 'S'
        else:
            chars[3] = 's'
    if mode & stat.S_ISGID:
        if chars[6] != 'x':
            chars[6] = 'S'
        else:
            chars[6] = 's'
    if mode & stat.S_ISVTX:
        if chars[9] != 'x':
            chars[9] = 'T'
        else:
            chars[9] = 't'
    return "%s (%o)" % (''.join(chars), mode)

