from hachoir.tools import makePrintable
from text_ui import TextOutput
import curses
import signal

class Display(TextOutput):
    def __init__(self, ui, *args):
        TextOutput.__init__(self, *args)
        self.ui = ui

    def writeLine(self, msg):
        self.ui.buffer.append(msg)
        if self.ui.leftmax < len(msg)-1:
            self.ui.leftmax = len(msg)-1
        height, width = self.ui.stdscr.getmaxyx()
        if self.ui.need_redraw:
            if height - 1 < len(self.ui.buffer):
                self.ui.need_redraw = False
                self.ui.scrollScreen(redraw=True)
                signal.signal(signal.SIGWINCH, self.ui.redrawWindow)
            else:
                self.ui.wrappedWrite(msg, width)
                self.ui.stdscr.refresh()

class CursesUI:
    def __init__(self, charset):
        self.term_charset = charset
        self.stdscr = None
        self.buffer = []
        self.parsing_done = False
        self.need_redraw = True
        self.top = 0
        self.left = 0
        self.leftmax = 0
        self.footer = "[ q to quit - move with arrows, page up/down, home/end ]"
        self.footer_waiting = "Hachoir is parsing your stream (be patient) ..."
        self.field_set_changed = False

    def wrappedWrite(self, text, width, newline=True):
        text = makePrintable(text, self.term_charset, to_unicode=True)
        add_newline = False
        if newline:
            if width <= (len(text)):
                text = text[:width-1]+"$"
            elif (len(text)+1) <= width:
                add_newline = True
        else:
            if len(text) < width:
                pass
            else:
                text = text[:width-2]+"$"
        if add_newline:
            text += "\n"
        text = text.encode(self.term_charset)
        self.stdscr.addstr(text)

    def redrawWindow(self, *unused_args):
        try:
            curses.endwin()
        except:
            pass
        self.initCurses()
        self.scrollScreen(redraw=True)

    def scrollScreen(self,
        top_delta=None, top_value=None,
        left_delta=None, left_value=None,
        redraw=False):

        # Get console size
        height, width = self.stdscr.getmaxyx()
        height -= 1

        oldleft, oldtop = self.left, self.top

        # Update left (if needed)
        if left_value is not None:
            self.left = left_value
        if left_delta is not None:
            self.left += left_delta
        if self.left < 0:
            self.left = 0
        elif self.leftmax < self.left:
            self.left = self.leftmax

        # Update top (if needed)
        if top_value is not None:
            self.top = top_value
        if top_delta is not None:
            self.top += top_delta
        if height < len(self.buffer):
            topmax = len(self.buffer)-height
        else:
            topmax = 0
        if self.top < 0:
            self.top = 0
        elif topmax < self.top:
            self.top = topmax

        if (self.left == oldleft and self.top == oldtop) and not redraw:
            return

        # Redraw screen
        self.stdscr.clear()
        for line in self.buffer[self.top:self.top+height]:
            self.wrappedWrite(line[self.left:], width)

        if self.parsing_done:
            text = self.footer
        else:
            text = self.footer_waiting
        self.wrappedWrite(text, width, newline=False)
        self.stdscr.refresh()

    def fieldSetChanged(self, *unused_args):
        self.field_set_changed = True

    def display(self, field_set, max_depth, options):
        KEY_LEFT = 260
        KEY_RIGHT = 261
        KEY_UP = 259
        KEY_DOWN = 258
        KEY_PAGEUP = 339
        KEY_PAGEDOWN = 338
        KEY_HOME = 262
        KEY_END = 360
        KEY_ESCAPPE = 27

        # Catch "field-replaced" event
        field_set.connectEvent("field-replaced", self.fieldSetChanged, local=False)
        field_set.connectEvent("field-insered", self.fieldSetChanged, local=False)

        # Parse and display output
        output = Display(self, field_set, self.term_charset, max_depth, options)
        output.display()
        self.parsing_done = True

        # Write footer
        self.scrollScreen(redraw=True)

        # Field set changed?
        if self.field_set_changed:
            self.field_set_changed = False
            self.buffer = []
            self.parsing_done = False
            self.need_redraw = True
            self.leftmax = 0
            self.footer_waiting = "Restart drawing since field has changed ..."
            self.stdscr.clear()
            output.display(self.cursesWrite)
            self.parsing_done = True
            self.scrollScreen(redraw=True)

        # Install our "resize terminal" signal handler
        signal.signal(signal.SIGWINCH, self.redrawWindow)

        # Keyboard main loop (interaction)
        while True:
            # Read a key
            keycode = self.stdscr.getch()

            # Key action ...
            height, width = self.stdscr.getmaxyx()
            if keycode in (ord("q"), ord("Q"), KEY_ESCAPPE):
                break
            if keycode == KEY_LEFT:
                self.scrollScreen(left_delta=-1)
            elif keycode == KEY_RIGHT:
                self.scrollScreen(left_delta=1)
            elif keycode == KEY_UP:
                self.scrollScreen(top_delta=-1)
            elif keycode == KEY_DOWN:
                self.scrollScreen(top_delta=1)
            elif keycode == KEY_PAGEUP:
                self.scrollScreen(top_delta=-height+1)
            elif keycode == KEY_PAGEDOWN:
                self.scrollScreen(top_delta=height-1)
            elif keycode == KEY_HOME:
                self.scrollScreen(top_value=0)
            elif keycode == KEY_END:
                self.scrollScreen(top_value=len(self.buffer)-height+1)
            elif keycode in (ord("r"), ord("R")):
                self.scrollScreen(redraw=True)

    def initCurses(self):
        self.stdscr = curses.initscr()
        self.stdscr.scrollok(True)
        curses.noecho()
        self.stdscr.keypad(1)

    def deinitCurses(self):
        self.stdscr.clear()
        curses.nocbreak()
        self.stdscr.keypad(0)
        curses.echo()
        curses.endwin()

def displayFieldSet(field_set, charset, max_depth, options):
    ui = CursesUI(charset)
    ui.initCurses()
    try:
        ui.display(field_set, max_depth, options)
    finally:
        ui.deinitCurses()

